import re
from dataclasses import dataclass
from json import JSONDecodeError
from typing import TYPE_CHECKING, Tuple, Union, List, Callable

import pandas as pd
import requests

from ._enums import CustomInstrumentTypes
from ._instrument_prop_classes import Serializable
from .._content_data_provider import ContentDataProvider
from .._historical_content_validator import HistoricalContentValidator
from .._historical_data_provider import (
    EventsDataProvider,
    SummariesDataProvider,
)
from .._historical_response_factory import HistoricalResponseFactory
from .._intervals import DayIntervalType, get_day_interval_type, Intervals
from ..historical_pricing._historical_pricing_request_factory import check_count
from ..ipa.dates_and_calendars.holidays._holidays_data_provider import Holiday
from ..._content_type import ContentType
from ..._core.session import get_valid_session
from ..._errors import RDError
from ..._tools import (
    get_response_reason,
    make_enum_arg_parser,
    custom_inst_datetime_adapter,
    ParamItem,
    ValueParamItem,
    EnumArgsParser,
    make_parse_enum,
    make_convert_to_enum,
    is_date_true,
    cached_property,
    extend_params,
)
from ..._tools._dataframe import convert_df_columns_to_datetime, convert_dtypes
from ...delivery._data._data_provider import (
    RequestFactory,
    Parser,
    success_http_codes,
    ContentValidator,
    ContentTypeValidator,
    ValidatorContainer,
    ParsedData,
)
from ...delivery._data._data_provider import _check_response as default_check_response
from ...delivery._data._data_provider_layer import get_data_by_data_type
from ...delivery._data._endpoint_data import RequestMethod

if TYPE_CHECKING:
    import httpx

content_type_by_day_interval_type = {
    DayIntervalType.INTER: ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES,
    DayIntervalType.INTRA: ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES,
}

# a20140be-3648-4892-9d1b-ce78ee8617fd
is_instrument_id = re.compile(r"[a-z0-9]{8}(-[a-z0-9]{4}){3}-[a-z0-9]{12}")

# S)INST.GESG1-0000
symbol_with_user_id = re.compile(r".*\.[A-Z0-9]+-[A-Z0-9]+")

wrong_uuid_regexp = re.compile(
    r"(Validation Error: .UUID suffix ).*( not matched with userID)"
)
wrong_symbol = "S)Instrument.UUID-0000"


def provide_session(func):
    def _func(value, session, *args, **kwargs):
        return func(value, session)

    return _func


def get_content_type_by_interval(interval) -> ContentType:
    day_interval_type = get_day_interval_type(interval)
    return content_type_by_day_interval_type.get(day_interval_type)


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


def custom_instruments_build_df(content_data: dict, **kwargs) -> pd.DataFrame:
    if isinstance(content_data, dict):
        content_data = [content_data]
    dataframe = pd.DataFrame(content_data)
    dataframe = convert_dtypes(dataframe)
    return dataframe


def custom_instruments_intervals_build_df(content_data: dict, **kwargs) -> pd.DataFrame:
    data = content_data.get("data")
    headers = content_data.get("headers", [])
    columns = [header.get("name") for header in headers]
    dataframe = pd.DataFrame(data, columns=columns)
    convert_df_columns_to_datetime(dataframe, entry="DATE", utc=True, delete_tz=True)
    dataframe.fillna(pd.NA, inplace=True)
    return dataframe


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------
def has_all_error_user_id(response):
    return all(wrong_uuid_regexp.match(error.message) for error in response.errors)


def check_response(response, config):
    return (
        None
        if has_all_error_user_id(response)
        else default_check_response(response, config)
    )


def get_user_id(session=None) -> str:
    session = get_valid_session(session)
    response = get_data_by_data_type(
        ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS, session, universe=wrong_symbol
    )
    check_response(response, session.config)
    errors = response.errors
    messages = [error.message for error in errors]
    user_id = ""
    for message in messages:
        if wrong_uuid_regexp.match(message):
            _, user_id = message.rsplit(" ", 1)
            break

    return user_id


def convert_to_symbol(symbol, session=None, uuid=""):
    # "MyNewInstrument"
    retval = symbol
    if not retval.startswith("S)"):
        retval = f"S){retval}"
    # "S)MyNewInstrument"
    if not symbol_with_user_id.match(retval):
        if not uuid:
            uuid = get_user_id(session)
        retval = f"{retval}.{uuid}"
    # "S)MyNewInstrument.GE-1234"
    return retval


def convert_to_holidays(array: List[Union[dict, Holiday]]) -> List:
    converted_holidays = []
    for holiday in array:
        if isinstance(holiday, dict):
            if "date" in holiday and "reason" in holiday:
                converted_holidays.append(holiday)
            else:
                raise ValueError("Holiday object should have 'date' and 'reason'")
        elif isinstance(holiday, Holiday):
            converted_holidays.append({"date": holiday.date, "reason": holiday.name})
        else:
            raise TypeError("holidays parameter can take only dict or Holiday objects")
    return converted_holidays


def convert_to_dict(obj: Union[dict, dataclass]) -> dict:
    if isinstance(obj, Serializable):
        return obj._to_dict()
    elif isinstance(obj, dict):
        return obj
    raise TypeError(f"Parameter can take only dict or UDC/Basket object")


def get_valid_symbol(symbol, uuid):
    return convert_to_symbol(symbol, uuid=uuid)


def get_valid_symbol_request(symbol, session):
    return convert_to_symbol(symbol, session)


class BaseRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        url = super().get_url(*args, **kwargs)
        if self.get_request_method(**kwargs) != RequestMethod.POST:
            url += "/{universe}"
        return url

    def get_path_parameters(self, session, *, universe=None, **kwargs):
        if self.get_request_method(**kwargs) == RequestMethod.POST:
            return {}

        if universe is None:
            raise RDError(-1, "universe can't be None")

        if not is_instrument_id.match(universe):
            universe = get_valid_symbol_request(universe, session)

        return {"universe": universe}

    def extend_query_parameters(self, query_parameters, extended_params=None):
        return extend_params(query_parameters, extended_params)

    def extend_body_parameters(self, body_parameters, **kwargs):
        return body_parameters


custom_inst_type_arg_parser = EnumArgsParser(
    parse=make_parse_enum(CustomInstrumentTypes),
    parse_to_enum=make_convert_to_enum(CustomInstrumentTypes),
)


class CustomInstsRequestFactory(BaseRequestFactory):
    @property
    def body_params_config(self):
        return custom_insts_body_params

    def get_body_parameters(self, session, *args, **kwargs):
        body_parameters = {}
        if self.get_request_method(**kwargs) not in {
            RequestMethod.POST,
            RequestMethod.PUT,
        }:
            return body_parameters

        return super().get_body_parameters(session, *args, **kwargs)

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            result = dict(body_parameters)
            result.update(extended_params)
            return result
        return body_parameters


# --------------------------------------------------------------------------------------
#   Raw data parser
# --------------------------------------------------------------------------------------


class CustomInstsParser(Parser):
    def parse_raw_response(
        self, raw_response: "httpx.Response"
    ) -> Tuple[bool, ParsedData]:
        is_success = False

        if raw_response is None:
            return is_success, ParsedData({}, {})

        is_success = raw_response.status_code in success_http_codes + [
            requests.codes.no_content
        ]

        if is_success:
            parsed_data = self.process_successful_response(raw_response)

        else:
            parsed_data = self.process_failed_response(raw_response)

        return is_success, parsed_data

    def process_failed_response(self, raw_response: "httpx.Response") -> ParsedData:
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }

        try:
            content_data = raw_response.json()
            if isinstance(content_data, list):
                content_data = content_data[0]
            content_error = content_data.get("error")

            if content_error:
                status["error"] = content_error
                error_code = content_error.get("code")
                if isinstance(error_code, str) and not error_code.isdigit():
                    error_code = raw_response.status_code
                error_message = content_error.get("message")
                errors = content_error.get("errors", {})
                errors = [error.get("reason") for error in errors if error]
                if errors:
                    errors = "\n".join(errors)
                    error_message = f"{error_message}: {errors}"
            elif "state" in content_data:
                state = content_data.get("state", {})
                error_code = state.get("code")
                data = content_data.get("data", [])
                reasons = [_data.get("reason", "") for _data in data]
                reason = "\n".join(reasons)
                error_message = f"{state.get('message')}: {reason}"
            else:
                error_code = raw_response.status_code
                error_message = raw_response.text

        except (TypeError, JSONDecodeError):
            error_code = raw_response.status_code
            error_message = raw_response.text

        if error_code == 403:
            if not error_message.endswith("."):
                error_message += ". "
            error_message += "Contact Refinitiv to check your permissions."

        return ParsedData(
            status, raw_response, error_codes=error_code, error_messages=error_message
        )


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class CustomInstsContentValidator(ContentValidator):
    @classmethod
    def content_data_is_not_none(cls, data: ParsedData) -> bool:
        if data.content_data is None and data.status.get("http_status_code") != 204:
            data.error_codes = 1
            data.error_messages = "Content data is None"
            return False
        return True

    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [self.content_data_is_not_none]


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------

interval_arg_parser = make_enum_arg_parser(Intervals, can_be_lower=True)


class CustomInstsSearchRequestFactory(RequestFactory):
    def get_query_parameters(self, *args, **kwargs):
        access = kwargs.get("access")
        return [
            ("access", access),
        ]

    def extend_query_parameters(self, query_parameters, extended_params=None):
        return extend_params(query_parameters, extended_params)

    def extend_body_parameters(self, body_parameters, **kwargs):
        return body_parameters


custom_insts_events_query_params = [
    ValueParamItem(
        "start", function=custom_inst_datetime_adapter.get_str, is_true=is_date_true
    ),
    ValueParamItem(
        "end", function=custom_inst_datetime_adapter.get_str, is_true=is_date_true
    ),
    ValueParamItem("count", function=check_count),
]
custom_insts_summaries_query_params = [
    ValueParamItem("interval", function=interval_arg_parser.get_str),
    ValueParamItem(
        "start", function=custom_inst_datetime_adapter.get_str, is_true=is_date_true
    ),
    ValueParamItem(
        "end", function=custom_inst_datetime_adapter.get_str, is_true=is_date_true
    ),
    ValueParamItem("count", function=check_count),
]

custom_insts_body_params = [
    ParamItem("exchange_name", "exchangeName"),
    ParamItem("instrument_name", "instrumentName"),
    ParamItem("time_zone", "timeZone"),
    ValueParamItem("type_", "type", custom_inst_type_arg_parser.parse),
    ParamItem("symbol", function=provide_session(convert_to_symbol)),
    ParamItem("currency"),
    ParamItem("description"),
    ParamItem("formula"),
    ValueParamItem("holidays", function=convert_to_holidays),
    ValueParamItem("basket", function=convert_to_dict),
    ValueParamItem("udc", function=convert_to_dict),
]

simple_custom_insts_body_params_config = [
    ParamItem("exchangeName"),
    ParamItem("instrumentName"),
    ParamItem("timeZone"),
    ParamItem("type"),
    ParamItem("symbol", function=provide_session(convert_to_symbol)),
    ParamItem("currency"),
    ParamItem("description"),
    ParamItem("formula"),
    ValueParamItem("udc", function=convert_to_dict),
    ValueParamItem("basket", function=convert_to_dict),
    ValueParamItem("holidays", function=convert_to_holidays),
]


class CustomInstsEventsRequestFactory(BaseRequestFactory):
    @property
    def query_params_config(self):
        return custom_insts_events_query_params


class CustomInstsSummariesRequestFactory(BaseRequestFactory):
    @property
    def query_params_config(self):
        return custom_insts_summaries_query_params


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------

custom_instrument_data_provider = ContentDataProvider(
    request=CustomInstsRequestFactory(),
    parser=CustomInstsParser(),
    validator=ValidatorContainer(
        content_validator=CustomInstsContentValidator(),
        content_type_validator=ContentTypeValidator({"application/json", ""}),
    ),
)

custom_instrument_search_data_provider = ContentDataProvider(
    request=CustomInstsSearchRequestFactory(),
    parser=CustomInstsParser(),
    validator=ValidatorContainer(
        content_validator=CustomInstsContentValidator(),
        content_type_validator=ContentTypeValidator({"application/json", ""}),
    ),
)

custom_instruments_events_data_provider = EventsDataProvider(
    request=CustomInstsEventsRequestFactory(),
    parser=CustomInstsParser(),
    response=HistoricalResponseFactory(),
    validator=HistoricalContentValidator(),
)

custom_instruments_intraday_summaries_data_provider = SummariesDataProvider(
    request=CustomInstsSummariesRequestFactory(),
    parser=CustomInstsParser(),
    response=HistoricalResponseFactory(),
    validator=HistoricalContentValidator(),
)

custom_instruments_interday_summaries_data_provider = SummariesDataProvider(
    request=CustomInstsSummariesRequestFactory(),
    parser=CustomInstsParser(),
    response=HistoricalResponseFactory(),
    validator=HistoricalContentValidator(),
)



from ._bid_ask import BidAsk
from .._types import OptFloat


class TenorBidAsk(BidAsk):
    """
    A tenor and corresponding swap bid, ask prices

    Parameters
    ----------
    tenor : str
        The code indicating the period between start date and end date
        of the instrument (e.g., '6M', '1Y').
    ask : float, optional
        Ask value
    bid : float, optional
        Bid value
    """

    def __init__(
        self,
        tenor: str,
        ask: OptFloat = None,
        bid: OptFloat = None,
    ) -> None:
        super().__init__(ask, bid)
        self.tenor = tenor

    @property
    def tenor(self):
        return self._get_param("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_param("tenor", value)



import json
from datetime import datetime

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from tests.integration.conftest import is_open
from tests.integration.constants_list import HttpStatusCode


def get_first_bond_definition():
    definition = rdf.bond.Definition(
        instrument_code="US1YT=RR",
        fields=[
            "InstrumentCode",
            "MarketDataDate",
            "YieldPercent",
            "GovernmentSpreadBp",
            "GovCountrySpreadBp",
            "RatingSpreadBp",
            "SectorRatingSpreadBp",
            "EdsfSpreadBp",
            "IssuerSpreadBp",
            "ErrorMessage",
        ],
    )
    return definition


def get_second_bond_definition():
    definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method="Dcb_Actual_Actual",
    )
    return definition


def get_third_bond_definition():
    definition = rdf.bond.Definition(
        instrument_code="US1YT=RR",
        fields=[
            "InstrumentCode",
            "MarketDataDate",
            "YieldPercent",
            "GovernmentSpreadBp",
            "GovCountrySpreadBp",
            "RatingSpreadBp",
            "SectorRatingSpreadBp",
            "EdsfSpreadBp",
            "IssuerSpreadBp",
        ],
        pricing_parameters=rdf.bond.PricingParameters(
            market_data_date="2019-07-05", price_side=rdf.bond.PriceSide.BID
        ),
    )
    return definition


def get_invalid_bond_definition():
    definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="INVALCCY",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method="Dcb_Actual_Actual",
    )
    return definition


def option_definition_01():
    definition = rdf.option.Definition(
        underlying_type=rdf.option.UnderlyingType.FX,
        strike=265,
        underlying_definition=rdf.option.FxUnderlyingDefinition("AUDUSD"),
        notional_ccy="AUD",
        tenor="5M",
        pricing_parameters=rdf.option.PricingParameters(
            price_side=rdf.option.PriceSide.MID,
            valuation_date="2018-08-06",
            pricing_model_type=rdf.option.PricingModelType.BLACK_SCHOLES,
            fx_spot_object=rdf.option.BidAskMid(
                bid=0.7387,
                ask=0.7387,
                mid=0.7387,
            ),
        ),
    )
    return definition


def financial_contracts_definition_with_fields():
    definition = rdf.Definitions(
        [
            get_first_bond_definition(),
            get_second_bond_definition(),
            get_third_bond_definition(),
        ],
        fields=[
            "InstrumentCode",
            "MarketDataDate",
            "YieldPercent",
            "GovernmentSpreadBp",
            "GovCountrySpreadBp",
            "RatingSpreadBp",
            "SectorRatingSpreadBp",
            "EdsfSpreadBp",
            "IssuerSpreadBp",
            "ErrorMessage",
            "INVALID_FIELD",
        ],
    )
    return definition


def financial_contracts_definition_without_fields():
    definition = rdf.Definitions(
        [
            get_second_bond_definition(),
            option_definition_01(),
        ]
    )

    return definition


def get_valid_and_invalid_financial_contracts_definition():
    definition = rdf.Definitions(
        [
            get_first_bond_definition(),
            get_second_bond_definition(),
            get_invalid_bond_definition(),
        ],
        fields=[
            "InstrumentCode",
            "MarketDataDate",
            "YieldPercent",
            "GovernmentSpreadBp",
            "GovCountrySpreadBp",
            "RatingSpreadBp",
            "SectorRatingSpreadBp",
            "EdsfSpreadBp",
            "IssuerSpreadBp",
            "ErrorMessage",
            "INVALID_FIELD",
        ],
    )
    return definition


@pytest.fixture(
    scope="function", params=[False, True], ids=["async_mode=False", "async_mode=True"]
)
def async_mode(request):
    return request.param


def check_http_status_is_success_and_df_value_not_empty(response):
    status = response.http_status
    df = response.data.df
    assert status["http_status_code"] == HttpStatusCode.TWO_HUNDRED, (
        f"Actual status code is {status['http_status_code']}, "
        f"Error: {response.http_status['error']}"
    )

    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"
    assert df.MarketDataDate is not None, f"{df.MarketDataDate}"


def check_http_status_bad_request_and_error_message(response):
    status = response.http_status
    message = response.error_message
    assert (
        status["http_status_code"] == HttpStatusCode.FOUR_HUNDRED
    ), f"Actual status code is {status['http_status_code']}"

    assert message == "Validation error", f"Message: {message}"


def check_stream_state_and_df_from_stream(stream, expected_value=None):
    df = stream.get_snapshot()
    if expected_value is not None:
        actual_value = df.InstrumentDescription[0]
        assert expected_value in actual_value, f"Actual value: {actual_value}"

    assert is_open(stream), f"Stream is not open"
    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert df.MarketDataDate is not None, f"{df.MarketDataDate}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"

    stream.close()


def check_http_status_is_success(response):
    status = response.http_status

    assert status["http_status_code"] == HttpStatusCode.TWO_HUNDRED, (
        f"Actual status code is {status['http_status_code']}, "
        f"Error: {response.http_status['error']}"
    )


def check_request_message_contains_expected_fields(response, expected_fields_names):
    request_byte = response.request_message.content
    request = json.loads(request_byte.decode("utf-8"))
    actual_list = request["fields"]
    actual_list_lowercase = list(map(lambda x: x.lower(), actual_list))
    for field_name in expected_fields_names:
        assert (
            field_name.lower() in actual_list_lowercase
        ), f"Expected field name '{field_name}' is not found in list of fields in request: {actual_list}"


def display_event(
    data, instrument_name, stream, event_type, triggered_events_list=None
):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append(event_type)
    print("----------------------------------------------------------")
    print(">>> {} - {} - at {}".format(event_type, data, current_time))


def add_call_backs(stream, triggered_events_list=None):
    stream.on_update(
        lambda data, instrument_name, item_stream: display_event(
            data,
            instrument_name,
            item_stream,
            "Update received for",
            triggered_events_list,
        )
    )
    stream.on_response(
        lambda data, instrument_name, item_stream: display_event(
            data,
            instrument_name,
            item_stream,
            "Response received for",
            triggered_events_list,
        )
    )
    stream.on_state(
        lambda event, item_stream: display_event(
            event, item_stream, "State received for", triggered_events_list
        )
    )



import allure
import pytest

import refinitiv.data.session.platform as platform
import tests.integration.conftest as conf
from refinitiv.data import OpenState
from refinitiv.data.errors import PlatformSessionError
from refinitiv.data.session import set_default, EventCode
from tests.integration.core.session.conftest import (
    assert_session_open_state,
    on_state,
    on_event,
)
from refinitiv.data.delivery import omm_stream
from tests.integration.helpers import (
    compare_list,
    check_event_code,
    check_event_message
)


@allure.suite("Session layer - Platform session")
@allure.feature("Session layer - Platform session")
@allure.severity(allure.severity_level.CRITICAL)
class TestPlatformSession:
    @allure.title(
        "Verify that session is not opened and error notification received using invalid credentials"
    )
    @pytest.mark.caseid("C37689844")
    def test_session_is_not_opened_using_invalid_credentials(self):
        session = None
        try:
            session = platform.Definition(
                app_key="INVALID",
                grant=platform.GrantPassword(username="USERNAME", password="PASSWORD"),
            ).get_session()

            session.open()
        except PlatformSessionError:
            pass

        assert session.open_state == OpenState.Closed

    @allure.title(
        "Verify that Platform session with valid credentials opens and check callbacks"
    )
    @pytest.mark.parametrize(
        "expected_events",
        [[
            (OpenState.Pending, "Session opening in progress"),
            (EventCode.SessionAuthenticationSuccess, "All is well"),
            (OpenState.Opened, "Session is opened"),
            (OpenState.Closed, "Session is closed"),
        ]],
    )
    @pytest.mark.caseid("C37689397")
    def test_platform_session_with_callbacks(self, expected_events):
        event_list = []
        session = platform.Definition(
            app_key=conf.desktop_app_key,
            grant=platform.GrantPassword(
                username=conf.edp.username, password=conf.edp.password
            ),
        ).get_session()
        set_default(session)
        session.on_event(
            lambda event, message, session: on_event(
                event, message, session, event_list
            )
        )
        session.on_state(
            lambda open_state, message, session: on_state(
                open_state, message, session, event_list
            )
        )
        session.open()

        assert_session_open_state(session)
        compare_list(event_list, expected_events)

    @allure.title("Verify using platform session with ContextManager")
    @pytest.mark.caseid("C44023028")
    @pytest.mark.smoke
    def test_platform_session_is_opened_and_is_closed_using_context_manager(self):
        session = platform.Definition(
            app_key=conf.desktop_app_key,
            grant=platform.GrantPassword(
                username=conf.edp.username, password=conf.edp.password
            ),
        ).get_session()
        with session:
            assert session.open_state == OpenState.Opened
        assert session.open_state == OpenState.Closed

    @allure.title("Verify using platform session with asynchronous ContextManager")
    @pytest.mark.caseid("C44023030")
    @pytest.mark.smoke
    async def test_platform_session_is_opened_and_is_closed_using_async_context_manager(
        self,
    ):
        session = platform.Definition(
            app_key=conf.desktop_app_key,
            grant=platform.GrantPassword(
                username=conf.edp.username, password=conf.edp.password
            ),
        ).get_session()
        async with session:
            assert session.open_state == OpenState.Opened
        assert session.open_state == OpenState.Closed

    @allure.title(
        "Open Platform session with callbacks and verify that all cnx events are received"
    )
    @pytest.mark.parametrize(
        "expected_events",
        [[(OpenState.Pending, "Session opening in progress"),
          (OpenState.Opened, "Session is opened"),
          (EventCode.StreamConnecting, "apis.streaming.pricing.endpoints.main"),
          (EventCode.StreamConnected, "apis.streaming.pricing.endpoints.main"),
          (OpenState.Closed, "Session is closed")]],
    )
    def test_platform_session_with_callbacks(self, expected_events):
        event_list = []
        session = platform.Definition(
            app_key=conf.desktop_app_key,
            grant=platform.GrantPassword(
                username=conf.edp.username, password=conf.edp.password
            ),
        ).get_session()
        set_default(session)
        session.on_event(
            lambda event, message, session: on_event(
                event, message["api_cfg"], session, event_list
            )
        )
        session.on_state(
            lambda open_state, message, session: on_state(
                open_state, message, session, event_list
            )
        )
        session.open()
        stream = omm_stream.Definition(name="GBP=", fields=["BID", "ASK"]).get_stream(session=session)
        stream.open()
        assert stream.open_state == OpenState.Opened, "Stream is not in opened state"
        stream.close()
        assert stream.open_state == OpenState.Closed, "Stream is not in closed state"

        assert_session_open_state(session)
        session.close()
        check_event_code(event_list, expected_events)
        check_event_message(event_list, expected_events)



# coding: utf8

from enum import Enum, unique


@unique
class PricingModelType(Enum):
    BINOMIAL = "Binomial"
    BLACK_SCHOLES = "BlackScholes"
    LOCAL_VOLATILITY = "LocalVolatility"
    TRINOMIAL = "Trinomial"
    VANNA_VOLGA = "VannaVolga"
    WHALEY = "Whaley"



from datetime import date, datetime, timedelta
from logging import Logger
from typing import Dict, List, Optional, Union, Tuple, Iterable

from pandas import DataFrame

from ._containers import (
    UniverseContainer,
    FieldsContainer,
    ADCDataContainer,
    HPDataContainer,
    CustInstDataContainer,
)
from ._history_provider_factory import get_history_provider
from ._intervals_consts import INTERVALS, EVENTS_INTERVALS
from .context_collection import get_hp_context, get_adc_context, get_cust_inst_context
from .._core.session import get_default, raise_if_closed
from .._tools import (
    DEBUG,
    fr_datetime_adapter,
)
from .._types import OptDateTime
from ..content import (
    custom_instruments,
    fundamental_and_reference,
    historical_pricing,
)
from ..content.fundamental_and_reference._data_grid_type import (
    get_data_grid_type_by_session,
)
from ..usage_collection._filter_types import FilterType
from ..usage_collection._logger import get_usage_logger
from ..usage_collection._utils import ModuleName


def get_history(
    universe: Union[str, Iterable[str]],
    fields: Union[str, Iterable[str], None] = None,
    interval: Optional[str] = None,
    start: "OptDateTime" = None,
    end: "OptDateTime" = None,
    adjustments: Optional[str] = None,
    count: Optional[int] = None,
    use_field_names_in_headers: bool = False,
    parameters: Union[str, dict, None] = None,
) -> DataFrame:
    """
    Retrieves the pricing history, as well as Fundamental and Reference data history.

    Parameters
    ----------
    universe: str | list
        Instruments to request
    fields: str | list, optional
        Fields to request
    interval: str, optional
        Date interval. Supported intervals are:
        tick, tas, taq, minute, 1min, 5min, 10min, 30min, 60min, hourly, 1h, daily,
        1d, 1D, 7D, 7d, weekly, 1W, monthly, 1M, quarterly, 3M, 6M, yearly, 1Y
    start: str or date or datetime or timedelta, optional
        The start date and timestamp of the requested history
    end: str or date or datetime or timedelta, optional
        The end date and timestamp of the requested history
    adjustments : str, optional
        Tells the system whether to apply or not apply CORAX (Corporate Actions)
        events or exchange/manual corrections or price and volume adjustment
        according to trade/quote qualifier summarization actions to historical time
        series data. Possible values are:
        exchangeCorrection, manualCorrection, CCH, CRE, RTS, RPO, unadjusted,
        qualifiers
    count : int, optional
        The maximum number of data points returned. Values range: 1 - 10000.
        Applies only to pricing fields.
    use_field_names_in_headers : bool, default False
        If True - returns field name as column headers for data instead of title
    parameters: str | dict, optional
        Single global parameter key=value or dictionary
        of global parameters to request.
        Applies only to TR fields.

    Returns
    -------
    pandas.DataFrame

     Examples
    --------
    >>> get_history(universe="GOOG.O")
    >>> get_history(universe="GOOG.O", fields="tr.Revenue", interval="1Y")
    >>> get_history(
    ...     universe="GOOG.O",
    ...     fields=["BID", "ASK", "tr.Revenue"],
    ...     interval="1Y",
    ...     start="2015-01-01",
    ...     end="2020-10-01",
    ... )
    """
    session = get_default()
    raise_if_closed(session)

    logger = session.logger()

    if interval is not None and interval not in INTERVALS:
        raise ValueError(
            f"Not supported interval value.\nSupported intervals are:"
            f"{list(INTERVALS.keys())}"
        )

    # Library usage logging
    get_usage_logger().log_func(
        name=f"{ModuleName.ACCESS}.get_history",
        func_path=f"{__name__}.get_history",
        kwargs=dict(
            universe=universe,
            fields=fields,
            interval=interval,
            start=start,
            end=end,
            count=count,
            adjustments=adjustments,
            parameters=parameters,
            use_field_names_in_headers=use_field_names_in_headers,
        ),
        desc={FilterType.SYNC, FilterType.LAYER_ACCESS},
    )

    universe = UniverseContainer(universe)
    fields = FieldsContainer(fields)
    data_grid_type = get_data_grid_type_by_session(session)
    hp = get_hp_context(data_grid_type, universe, fields)
    adc = get_adc_context(data_grid_type, universe, fields)
    cust_inst = get_cust_inst_context(data_grid_type, universe, fields)

    adc_raw = None
    adc_df = None
    if adc.can_get_data:
        adc_params = get_adc_params(start, end, interval)
        adc_params.update(parameters or {})
        adc_raw, adc_df = get_adc_data(
            universe=universe.adc,
            fields=fields.adc,
            parameters=adc_params,
            use_field_names_in_headers=use_field_names_in_headers,
            logger=logger,
        )

    adc_data = ADCDataContainer(adc_raw, adc_df, fields)
    universe.calc_hp(adc_data.raw)

    hp_raw = None
    hp_df = None
    if hp.can_get_data:
        hp_raw, hp_df = get_hp_data(
            universe=universe.hp,
            interval=interval,
            start=start,
            end=end,
            adjustments=adjustments,
            count=count,
            fields=fields.hp,
            logger=logger,
        )

    hp_data = HPDataContainer(hp_raw, hp_df)
    cust_inst_raw = None
    cust_inst_df = None
    if cust_inst.can_get_data:
        cust_inst_raw, cust_inst_df = get_custominsts_data(
            universe=universe.cust_inst,
            interval=interval,
            start=start,
            end=end,
            count=count,
            logger=logger,
        )
    cust_inst_data = CustInstDataContainer(cust_inst_raw, cust_inst_df)

    if not any({adc_data, hp_data, cust_inst_data}):
        return DataFrame()

    adc.set_data(adc_data, hp_data, cust_inst_data)
    hp.set_data(adc_data, hp_data, cust_inst_data)
    cust_inst.set_data(adc_data, hp_data, cust_inst_data)

    history_provider = get_history_provider(data_grid_type)
    return history_provider.get_df(
        adc, hp, cust_inst, universe, fields, interval, use_field_names_in_headers
    )


def get_hp_data(
    universe: List[str],
    fields: List[str],
    interval: Optional[str],
    start: Optional[str],
    end: Optional[str],
    adjustments: Optional[str],
    count: Optional[int],
    logger: Logger,
):
    """Get historical pricing raw data.

    Args:
        universe (List[str]): Sequence of RICs.
        fields (List[str]): List of fields for request.
        interval (Optional[str]): consolidation interval.
        start (Optional[str]): start date.
        end (Optional[str]): end date.
        adjustments (Optional[str]): adjustments for request.
        count (Optional[int]): the maximum number of data returned.
        logger (Logger): session logger.
    """
    if interval in EVENTS_INTERVALS:
        definition = historical_pricing.events.Definition(
            universe=universe,
            eventTypes=INTERVALS[interval]["event_types"],
            start=start,
            end=end,
            adjustments=adjustments,
            count=count,
            fields=fields,
        )

    else:
        interval = INTERVALS[interval]["pricing"] if interval is not None else interval

        definition = historical_pricing.summaries.Definition(
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            adjustments=adjustments,
            count=count,
            fields=fields,
        )

    try:
        response = definition.get_data()
        DEBUG and logger.debug(
            f"HISTORICAL_PRICING --->\n{response.data.df.to_string()}\n"
        )
        raw = response.data.raw
        df = response.data.df
        return raw, df
    except Exception as e:
        if DEBUG:
            logger.exception(f"Failure sending request with {definition}, error:{e}")
        df = DataFrame()
        return {}, df


def get_adc_params(
    start: Union[str, date, datetime, timedelta],
    end: Union[str, date, datetime, timedelta],
    interval: Optional[str],
) -> dict:
    """Get parameters for ADC request.

    Args:
        start: start date for calculation parameters.
        end: end date for calculation parameters.
        interval: consolidated interval for calculation parameters.

    Returns:
        parameters: parameters for ADC request.
    """
    parameters = {}
    if start is not None:
        parameters["SDate"] = fr_datetime_adapter.get_str(start)

    if end is not None:
        parameters["EDate"] = fr_datetime_adapter.get_str(end)

    if interval is not None:
        parameters["FRQ"] = INTERVALS[interval]["adc"]

    return parameters


def get_adc_data(
    universe: List[str],
    fields: List[str],
    parameters: dict,
    use_field_names_in_headers: bool,
    logger: Logger,
) -> Union[Tuple[dict, DataFrame], Tuple[None, DataFrame]]:
    """Get ADC raw data.

    Args:
        universe (List[str]): sequence of RICs.
        fields (List[str]): list of fields for request.
        parameters (Dict[str]): precalculated parameters for request.
        use_field_names_in_headers (bool): return fields names in headers instead of title.
        logger (Logger): session logger.
    """
    definition = fundamental_and_reference.Definition(
        universe=universe,
        fields=fields,
        parameters=parameters,
        row_headers="date",
        use_field_names_in_headers=use_field_names_in_headers,
    )
    try:
        response = definition.get_data()
        raw = response.data.raw
        df = response.data.df
        DEBUG and logger.debug(f"ADC --->\n{response.data.df.to_string()}\n")
        return raw, df
    except Exception as e:
        if DEBUG:
            logger.exception(f"Failure sending request with {definition}. {e}")
        return {}, DataFrame()


def get_custominsts_data(
    universe: List[str],
    interval: Optional[str],
    start: Optional[str],
    end: Optional[str],
    count: Optional[int],
    logger: Logger,
) -> Union[Tuple[dict, DataFrame], Tuple[None, DataFrame]]:
    """Get custom instruments raw data.

    Args:
        universe (List[str]): list of RICs.
        interval (Optional[str]): optional interval.
        start (Optional[str]): optional start date.
        end (Optional[str]): optional end date.
        count (OptDict[int]): maximim number of retrieved data.
        logger (Logger): session logger.
    """
    if interval in EVENTS_INTERVALS:
        definition = custom_instruments.events.Definition(
            universe=universe,
            start=start,
            end=end,
            count=count,
        )

    else:
        interval = INTERVALS[interval]["pricing"] if interval is not None else interval
        definition = custom_instruments.summaries.Definition(
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            count=count,
        )

    try:
        response = definition.get_data()
        raw = response.data.raw
        df = response.data.df
        DEBUG and logger.debug(f"CUSTOMINSTS --->\n{response.data.df.to_string()}\n")
        return raw, df
    except Exception:
        if DEBUG:
            logger.exception(f"Failure sending request with {definition}")
        return {}, DataFrame()



import urllib.parse
from typing import Any, TYPE_CHECKING

from ._endpoint_data import RequestMethod
from ._request import Request
from ..._tools import parse_url, urljoin, bool_or_none_to_str, get_params

if TYPE_CHECKING:
    from ..._core.session import Session


class RequestFactory:
    def get_url(self, session: "Session", url: str, *args, **kwargs) -> str:
        return url

    def create(self, session, url, *args, **kwargs) -> Request:
        method = self.get_request_method(**kwargs)
        session.verify_scope(url, method)
        url = self.get_url(session, url, *args, **kwargs)
        url_root = session._get_rdp_url_root()

        header_parameters = self.get_header_parameters(session, **kwargs)
        path_parameters = self.get_path_parameters(session, **kwargs)
        query_parameters = self.get_query_parameters(*args, **kwargs)
        query_parameters = self.extend_query_parameters(
            query_parameters, extended_params=kwargs.get("extended_params")
        )
        closure = kwargs.get("closure")

        url = self.update_url(url_root, url, path_parameters, query_parameters)

        body_parameters = None
        headers = header_parameters
        if method != RequestMethod.GET:
            headers["Content-Type"] = "application/json"
        if method in (RequestMethod.POST, RequestMethod.PUT):
            body_parameters = self.get_body_parameters(session, *args, **kwargs)
            body_parameters = self.extend_body_parameters(body_parameters, **kwargs)

        request = Request(
            url=url,
            method=method,
            headers=headers,
            json=body_parameters,
        )

        if closure is not None:
            request.closure = closure

        return request

    def update_url(self, url_root, url, path_parameters, query_parameters):
        try:
            result = parse_url(url)

            if not all([result.scheme, result.netloc, result.path]):
                if result.query and result.path:
                    url = urljoin(url_root, result.path)
                else:
                    url = urljoin(url_root, url)
            else:
                url = "".join([url_root, result.path])

            if result.query and isinstance(query_parameters, list):
                query_parameters.extend(urllib.parse.parse_qsl(result.query))

        except Exception:
            url = "/".join([url_root + "/" + url])

        if path_parameters:
            for key, value in path_parameters.items():
                value = bool_or_none_to_str(value)
                value = urllib.parse.quote(str(value))
                url = url.replace("{" + key + "}", value)

        if self.is_multiple_query_parameters(query_parameters):
            url = self.add_multiple_query_parameters(url, query_parameters)
        elif query_parameters:
            url = self.add_query_parameters(url, query_parameters)

        return url

    def is_multiple_query_parameters(self, query_parameters: Any) -> bool:
        """Check multiple query parameters.

        Args:
            query_parameters (Any): Query parameters.

        Uses to check query parameters
        if they passes like {"universe": ["IBM.N", "ORCL.N"]}
        in case
        when query string must be constructed like '?universe=BNPP.PA,ORCL.N'
        """
        if isinstance(query_parameters, dict):
            return all(
                [
                    len(query_parameters) == 1,
                    isinstance(next(iter(query_parameters.values())), list),
                ]
            )

        return False

    def get_request_method(self, *, method=None, **kwargs) -> RequestMethod:
        return method or RequestMethod.GET

    @property
    def body_params_config(self):
        return []

    def get_body_parameters(self, *args, body_params_config=None, **kwargs) -> dict:
        body_params_config = body_params_config or self.body_params_config
        return dict(get_params(body_params_config, *args, **kwargs))

    @property
    def query_params_config(self):
        return []

    def get_query_parameters(self, *args, **kwargs) -> list:
        return get_params(self.query_params_config, *args, **kwargs)

    def get_path_parameters(
        self, session=None, *, path_parameters=None, **kwargs
    ) -> dict:
        return path_parameters or {}

    def add_multiple_query_parameters(self, url: str, query_parameters: dict) -> str:
        """Add multiple query parameters to query string.

        Args:
            url (str): url to construct query string.
            query_parameters (dict): query parameters to construct query string.

        Uses to construct percent-encoded query string
        if query parameters passed like {"universe": ["IBM.N", "ORCL.N"]}
        and
        must be transformed to query string like '?universe=BNPP.PA,ORCL.N',
        not like '?universe=BNPP.PA&universe=ORCL.N'

        Estimates API accepts only '?universe=BNPP.PA,ORCL.N' format to get estimates
        actual KPI measures for multiple universe.
        """
        key, value = query_parameters.popitem()
        return f"{url}?{urllib.parse.urlencode({key: ', '.join(value)})}"

    def add_query_parameters(self, url: str, query_parameters: dict) -> str:
        return f"{url}?{urllib.parse.urlencode(query_parameters)}"

    def extend_query_parameters(self, query_parameters, extended_params=None):
        return query_parameters

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if extended_params:
            body_parameters.update(extended_params)
        return body_parameters

    def get_header_parameters(self, session=None, header_parameters=None, **kwargs):
        return header_parameters or {}



from datetime import datetime

from refinitiv.data.content import trade_data_service


def display_event(event, stream, event_type, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append(event_type)
    print("----------------------------------------------------------")
    print(">>> {} event received at {}".format(event_type, current_time))
    print(event)


def display_event_complete(item_stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append("Complete")
    print("----------------------------------------------------------")
    print(">>> {} event received at {}".format("Complete", current_time))


def create_trade_data_service_with_all_callbacks(
    universe=None,
    api=None,
    fields=None,
    universe_type=trade_data_service.UniverseTypes.UserID,
    extended_params=None,
    triggered_events_list=None,
    session=None,
):
    trade_data_service_stream = (
        trade_data_service.Definition(
            universe=universe,
            api=api,
            fields=fields,
            finalized_orders=trade_data_service.FinalizedOrders.P1D,
            events=trade_data_service.Events.Full,
            universe_type=universe_type,
            extended_params=extended_params,
        )
        .get_stream(session=session)
        .on_update(
            lambda event, stream: display_event(event, "Update", triggered_events_list)
        )
        .on_add(
            lambda event, item_stream: display_event(
                event, item_stream, "Add", triggered_events_list
            )
        )
        .on_state(
            lambda event, item_stream: display_event(
                event, item_stream, "State", triggered_events_list
            )
        )
        .on_complete(
            lambda item_stream: display_event_complete(
                item_stream, triggered_events_list
            )
        )
    )

    return trade_data_service_stream


def check_triggered_event_list(triggered_events_list):
    expected_events_list = [
        "Complete",
        "State",
    ]
    for event in expected_events_list:
        assert (
            event in triggered_events_list
        ), f"found event '{triggered_events_list}' not handled by handlers or not triggered"



# coding: utf8

from enum import Enum, unique


@unique
class YieldType(Enum):
    ANNUAL_EQUIVALENT = "Annual_Equivalent"
    BOND_ACTUAL_364 = "Bond_Actual_364"
    BRAESS_FANGMEYER = "Braess_Fangmeyer"
    DISCOUNT_ACTUAL_360 = "Discount_Actual_360"
    DISCOUNT_ACTUAL_365 = "Discount_Actual_365"
    EUROLAND = "Euroland"
    ISMA = "Isma"
    JAPANESE_COMPOUNDED = "Japanese_Compounded"
    JAPANESE_SIMPLE = "Japanese_Simple"
    MONEY_MARKET_ACTUAL_360 = "MoneyMarket_Actual_360"
    MONEY_MARKET_ACTUAL_365 = "MoneyMarket_Actual_365"
    MONEY_MARKET_ACTUAL_ACTUAL = "MoneyMarket_Actual_Actual"
    MOOSMUELLER = "Moosmueller"
    NATIVE = "Native"
    QUARTERLY_EQUIVALENT = "Quarterly_Equivalent"
    SEMIANNUAL_EQUIVALENT = "Semiannual_Equivalent"
    TURKISH_COMPOUNDED = "TurkishCompounded"
    US_GOVT = "UsGovt"
    US_T_BILLS = "UsTBills"
    WEEKEND = "Weekend"



# coding: utf8

from enum import Enum, unique


@unique
class TimeStamp(Enum):
    CLOSE = "Close"
    CLOSE_LONDON5_PM = "CloseLondon5PM"
    CLOSE_NEW_YORK5_PM = "CloseNewYork5PM"
    CLOSE_TOKYO5_PM = "CloseTokyo5PM"
    DEFAULT = "Default"
    OPEN = "Open"
    SETTLE = "Settle"



import pytest

from refinitiv.data.content.symbol_conversion import AssetClass
from refinitiv.data.content.symbol_conversion import AssetState
from refinitiv.data.content.symbol_conversion._asset_class import (
    _transform_to_string,
    search_all_category_by_asset_class,
    rcsasset_category_genealogy_by_asset_class,
    create_asset_class_request_strings,
)
from refinitiv.data.content.symbol_conversion._definition import _prepare_filter


def test_symbology_transform_to_string_search_all_category():
    # when
    result = _transform_to_string(
        values=[AssetClass.COMMODITIES, AssetClass.EQUITIES],
        category=search_all_category_by_asset_class,
    )

    # then
    assert result == "'Commodities' 'Equities'"


def test_symbology_transform_to_string_search_all_category_rcsasset_category():
    # when
    result = _transform_to_string(
        values=[AssetClass.WARRANTS, AssetClass.CERTIFICATES],
        category=rcsasset_category_genealogy_by_asset_class,
    )

    # then
    assert result == "'A:AA' 'A:6N'"


@pytest.mark.parametrize(
    ("input_data", "expected_result"),
    (
        (
            [AssetClass.COMMODITIES, AssetClass.EQUITIES],
            ("SearchAllCategoryv3 in ('Commodities' 'Equities')", ""),
        ),
        (
            [AssetClass.COMMODITIES, AssetClass.WARRANTS],
            (
                "SearchAllCategoryv3 in ('Commodities')",
                "RCSAssetCategoryGenealogy in ('A:AA')",
            ),
        ),
        ([AssetClass.CERTIFICATES], ("", "RCSAssetCategoryGenealogy in ('A:6N')")),
    ),
)
def test_symbology_create_asset_class_request_string(input_data, expected_result):
    # when
    result = create_asset_class_request_strings(asset_class=input_data)

    # then
    assert result == expected_result


@pytest.mark.parametrize(
    ("input_data", "expected_result"),
    (
        (
            [AssetState.ACTIVE, [AssetClass.COMMODITIES, AssetClass.EQUITIES]],
            "AssetState eq 'AC' and (SearchAllCategoryv3 in ('Commodities' 'Equities'))",
        ),
        (
            [None, AssetClass.COMMODITIES],
            ("AssetState eq 'AC' and (SearchAllCategoryv3 in ('Commodities'))"),
        ),
        ([AssetState.INACTIVE, None], ("(AssetState ne 'AC' and AssetState ne null)")),
    ),
)
def test_prepare_filter(input_data, expected_result):
    # when
    result = _prepare_filter(*input_data)

    # then
    assert result == expected_result



__all__ = "ContentUsageLoggerMixin"

from types import SimpleNamespace
from typing import List, Callable

import pandas as pd

from ._content_data import Data
from ..delivery._data._parsed_data import ParsedData
from ..delivery._data._response import Response


def join_responses(
    responses: List[Response],
    join_dataframes: Callable = pd.concat,
    response_class=Response,
    data_class=Data,
    reset_index=False,
    limit: int = None,
) -> Response:
    def build_df(*args, **kwargs):
        dfs = []
        df = None

        for response in responses:
            dfs.append(response.data.df)

        all_dfs_is_none = all(a is None for a in dfs)
        if not all_dfs_is_none:
            df = join_dataframes(dfs)

        if reset_index and df is not None:
            df = df.reset_index(drop=True)
        if limit:
            df = df[:limit]
        return df

    if len(responses) == 1:
        return responses[0]

    raws = []
    http_statuses = []
    http_headers = []
    request_messages = []
    http_responses = []
    errors = []
    is_successes = []

    for response in responses:
        raws.append(response.data.raw)
        http_statuses.append(response.http_status)
        http_headers.append(response.http_headers)
        request_messages.append(response.request_message)
        http_responses.append(response.http_response)
        is_successes.append(response.is_success)

        if response.errors:
            errors += response.errors

    raw_response = SimpleNamespace()
    raw_response.headers = http_headers
    raw_response.request = request_messages
    is_success = any(is_successes)
    data_factory = SimpleNamespace()
    response = response_class(
        is_success, ParsedData(http_statuses, raw_response), data_factory
    )
    data_factory.create_data = lambda *args, **kwargs: data_class(
        raws, build_df, owner_=response
    )
    response.errors += errors
    response.http_response = http_responses

    return response



import refinitiv.data as rd
from tests.unit.conftest import StubSession, StubResponse


def test_df_not_empty_if_date_is_none():
    # fmt: off
    response = StubResponse({
            "links": {"count": 2}, "variability": "", "universe": [{
            "Instrument": "IBM",
            "Company Common Name": "International Business Machines Corp",
            "Organization PermID": "4295904307",
            "Reporting Currency": "USD"
        }],
            "data": [["IBM", "2023-01-18T00:00:00", None, "USD", 60210414250],
                     ["IBM", "2021-12-31T00:00:00", "2021-12-31T00:00:00", None, None]],
            "messages": {
                "codes": [[-1, -1, -2, -1, -1], [-1, -1, -1, -2, -2]],
                "descriptions": [{"code": -2, "description": "empty"},
                                 {"code": -1, "description": "ok"}]
            }, "headers": [{
            "name": "instrument", "title": "Instrument", "type": "string",
        }, {
            "name": "date", "title": "Date", "type": "datetime",
        }, {
            "name": "TR.Revenue", "title": "Date", "type": "datetime",
        }, {
            "name": "TR.RevenueMean", "title": "Currency", "type": "string",
        }, {
            "name": "TR.RevenueMean", "title": "Revenue - Mean",
            "type": "number", "decimalChar": ".",
        }]})
    # fmt: on
    session = StubSession(is_open=True, response=response)
    session.config.set_param("apis.data.datagrid.underlying-platform", "rdp")
    rd.session.set_default(session)

    df = rd.get_history(
        universe="IBM",
        fields=["TR.Revenue.date", "TR.RevenueMean.currency", "TR.RevenueMean"],
        interval="daily",
        use_field_names_in_headers=True,
    )

    rd.session.set_default(None)

    assert not df.empty



import refinitiv.data.content.ipa.financial_contracts as rdf

cds_universe = {
    "instrumentType": "Cds",
    "instrumentDefinition": {
        "instrumentTag": "Cds1_InstrumentCode",
        "instrumentCode": "BNPP5YEUAM=R",
        "cdsConvention": "ISDA",
        "endDateMovingConvention": "NoMoving",
        "adjustToIsdaEndDate": True,
    },
    "pricingParameters": {"marketDataDate": "2020-01-01"},
}


def cds_definition():
    definition = rdf.cds.Definition(
        instrument_tag="Cds1_InstrumentCode",
        instrument_code="BNPP5YEUAM=R",
        cds_convention=rdf.cds.CdsConvention.ISDA,
        end_date_moving_convention=rdf.cds.BusinessDayConvention.NO_MOVING,
        step_in_date="2020-01-02",
        adjust_to_isda_end_date=True,
        pricing_parameters=rdf.cds.PricingParameters(market_data_date="2020-01-01"),
        fields=[
            "InstrumentTag",
            "ValuationDate",
            "InstrumentDescription",
            "StartDate",
            "EndDate",
            "SettlementDate",
            "UpfrontAmountInDealCcy",
            "CashAmountInDealCcy",
            "AccruedAmountInDealCcy",
            "AccruedBeginDate",
            "NextCouponDate",
            "UpfrontPercent",
            "ConventionalSpreadBp",
            "ParSpreadBp",
            "AccruedDays",
            "ErrorCode",
            "ErrorMessage",
            "MarketDataDate",
        ],
    )
    return definition



from refinitiv.data import _configure as configure


def test_get_config_default_user(write_user_config):
    write_user_config('{"user_prop": "user_value"}')

    configure.reload()

    assert configure.get_config().get("user_prop") == "user_value"


def test_get_config_default_project(write_project_config):
    write_project_config('{"project_prop": "project_value"}')

    configure.reload()

    assert configure.get_config().get("project_prop") == "project_value"


def test_get_config_default_user_project(write_user_config, write_project_config):
    write_project_config('{"project_prop": "project_value"}')
    write_user_config('{"user_prop": "user_value"}')

    configure.reload()

    assert configure.get_config().get("user_prop") == "user_value"
    assert configure.get_config().get("project_prop") == "project_value"


def test_get_config_override_default_prop_by_user(write_user_config):
    default_prop = "endpoint.env"
    user_value = "user_value"
    write_user_config('{"%s": "%s"}' % (default_prop, user_value))

    configure.reload()
    assert configure.get_config().get(default_prop) == user_value


def test_get_config_override_default_prop_by_project(write_project_config):
    default_prop = "endpoint.env"
    project_value = "project_value"
    write_project_config('{"%s": "%s"}' % (default_prop, project_value))

    configure.reload()
    assert configure.get_config().get(default_prop) == project_value


def test_get_config_override_user_prop_by_project(
    write_user_config, write_project_config
):
    prop = "endpoint.env"
    project_value = "project_value"
    user_value = "user_value"
    write_project_config('{"%s": "%s"}' % (prop, project_value))
    write_user_config('{"%s": "%s"}' % (prop, user_value))

    configure.reload()
    assert configure.get_config().get(prop) == project_value


def test_get_config_does_not_override_project_prop_by_user(
    write_user_config, write_project_config
):
    default_prop = "endpoint.env"
    project_value = "project_value"
    user_value = "user_value"
    write_project_config('{"%s": "%s"}' % (default_prop, project_value))
    write_user_config('{"%s": "%s"}' % (default_prop, user_value))

    configure.reload()
    value = configure.get_config().get(default_prop)
    assert value != user_value and value == project_value



import pytest

from refinitiv.data._tools._common import (
    ArgsParser,
    parse_list_of_str,
    fields_arg_parser,
    make_enum_arg_parser,
    make_parse_enum,
)
from .conftest import StubStrEnum


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (None, [None]),
        ([None], [None]),
        ("", [""]),
        ({}, [{}]),
        (tuple(), [()]),
        (set(), [set()]),
        ([], []),
    ],
)
def test_args_parser_get_list_return_list_type(input_value, expected_value):
    # given
    parser = ArgsParser(lambda _: _)

    # when
    l = parser.get_list(input_value)

    # then
    assert l == expected_value


def test_args_parser_get_str_return_str_type():
    # given
    parser = ArgsParser(lambda _: _)

    # when
    s = parser.get_str(None)

    # then
    assert isinstance(s, str)


def test_args_parser_get_str_with_delim_return_str_without_delim_if_only_one():
    # given
    delim = ","
    parser = ArgsParser(lambda _: _)

    # when
    s = parser.get_str([1], delim=delim)

    # then
    assert delim not in s


def test_args_parser_get_str_with_delim_return_str_with_delim_if_many():
    # given
    delim = ","
    parser = ArgsParser(lambda _: _)

    # when
    s = parser.get_str([1, 2], delim=delim)

    # then
    assert delim in s, s


@pytest.mark.parametrize(
    "input_value, expected_value",
    [([], []), ("string", ["string"]), (["string"], ["string"])],
)
def test_parse_list_of_str_success(input_value, expected_value):
    # when
    testing_value = parse_list_of_str(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value",
    [
        [None],
        [1],
        ["string", 1],
    ],
)
def test_parse_list_of_str_raise_error_not_all_strings(input_value):
    # then
    with pytest.raises(ValueError, match="Not all elements are strings"):
        # when
        parse_list_of_str(input_value)


@pytest.mark.parametrize(
    "input_value",
    [
        None,
        1,
        {},
        set(),
        tuple(),
    ],
)
def test_parse_list_of_str_raise_error_invalid_type(input_value):
    # then
    with pytest.raises(TypeError, match="Invalid type, expected str or list"):
        # when
        parse_list_of_str(input_value)


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (StubStrEnum.VALUE_1, "VALUE_1"),
        (StubStrEnum.VALUE_1.value, "VALUE_1"),
        ("VALUE_1", "VALUE_1"),
        ([StubStrEnum.VALUE_1, "VALUE_2"], ["VALUE_1", "VALUE_2"]),
    ],
)
def test_make_parse_enum_success(input_value, expected_value):
    # given
    parse_enum = make_parse_enum(StubStrEnum)

    # when
    testing_value = parse_enum(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (StubStrEnum.VALUE_1, "VALUE_1"),
        (StubStrEnum.VALUE_1.value, "VALUE_1"),
        ("VALUE_1", "VALUE_1"),
        ([StubStrEnum.VALUE_1, "VALUE_2"], ["VALUE_1", "VALUE_2"]),
        (StubStrEnum.VALUE_1.value.lower(), "VALUE_1"),
        ("value_1", "VALUE_1"),
        ([StubStrEnum.VALUE_1.value.lower(), "value_2"], ["VALUE_1", "VALUE_2"]),
    ],
)
def test_make_parse_enum_with_can_be_lower_success(input_value, expected_value):
    # given
    parse_enum = make_parse_enum(StubStrEnum, can_be_lower=True)

    # when
    testing_value = parse_enum(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value",
    [
        None,
        1,
        {},
        "",
        "None",
    ],
)
def test_make_parse_enum_success_raise_error(input_value):
    # given
    parse_enum = make_parse_enum(StubStrEnum)

    # then
    with pytest.raises(AttributeError, match=f"Value '{input_value}' must be in"):
        # when
        parse_enum(input_value)


def test_fields_arg_parser():
    # given, when, then
    assert fields_arg_parser


def test_make_enum_arg_parser():
    # when
    parser = make_enum_arg_parser(StubStrEnum)

    # then
    assert parser


def test_args_parser_get_unique_return_list_type():
    # given
    parser = ArgsParser(lambda _: _)

    # when
    testing_value = parser.get_unique(None)

    # then
    assert isinstance(testing_value, list)


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        ([1, 1, 1, 2, 2, 3], [1, 2, 3]),
        (["B", "C", "B", "A", "D"], ["B", "C", "A", "D"]),
        ([], []),
        (None, [None]),
    ],
)
def test_args_parser_get_unique_return_unique_list(input_value, expected_value):
    # given
    parser = ArgsParser(lambda _: _)

    # when
    testing_value = parser.get_unique(input_value)

    # then
    assert testing_value == expected_value



from inspect import signature

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession
from tests.unit.conftest import remove_dunder_methods, remove_private_attributes
from tests.unit.content.ipa.contracts.conftest import (
    has_property_names_in_class,
    get_property_names,
)


@pytest.mark.parametrize(
    argnames="input_data",
    ids=[
        "Definition",
        "RepoParameters",
        "PricingParameters",
        "UnderlyingContract",
        "UnderlyingPricingParameters",
    ],
    argvalues=[
        (
            rdf.repo._repo_definition.RepoInstrumentDefinition,
            {
                "instrument_tag": "instrument_tag",
                "start_date": "start_date",
                "end_date": "end_date",
                "tenor": "tenor",
                "day_count_basis": rdf.repo.DayCountBasis.DCB_30_365_BRAZIL,
                "underlying_instruments": [rdf.repo.UnderlyingContract()],
                "is_coupon_exchanged": True,
                "repo_rate_percent": 8.8,
                "buy_sell": rdf.repo.BuySell.BUY,
            },
        ),
        (
            rdf.repo.RepoParameters,
            {
                "coupon_paid_at_horizon": True,
                "haircut_rate_percent": 2.2,
                "initial_margin_percent": 3.3,
                "repurchase_price": 4.4,
                "purchase_price": 5.5,
            },
        ),
        (
            rdf.repo.PricingParameters,
            {
                "market_data_date": "market_data_date",
                "repo_curve_type": rdf.repo.RepoCurveType.LIBOR_FIXING,
                "valuation_date": "valuation_date",
                "report_ccy": "report_ccy",
            },
        ),
        (
            rdf.repo.UnderlyingContract,
            {
                "instrument_type": "instrument_type",
                "instrument_definition": rdf.bond.Definition(),
                "pricing_parameters": rdf.repo.UnderlyingPricingParameters(),
            },
        ),
        (
            rdf.repo.UnderlyingPricingParameters,
            {
                "pricing_parameters_at_end": rdf.bond.PricingParameters(),
                "pricing_parameters_at_start": rdf.bond.PricingParameters(),
                "repo_parameters": rdf.repo.RepoParameters(),
                "valuation_date": "valuation_date",
                "market_data_date": "market_data_date",
                "report_ccy": "report_ccy",
            },
        ),
    ],
)
def test_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    # +1 for (self)
    assert len(s.parameters) == (
        len(args_names) + 1
    ), f"csl={cls}{str(set(get_property_names(cls)) - set(args_names))}"

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k


def test_ipa_repo_definition_attributes():
    expected_attributes = ["get_data", "get_data_async", "get_stream"]
    testing_attributes = dir(rdf.repo.Definition)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


def test_ipa_repo_attributes():
    expected_attributes = [
        "BuySell",
        "DayCountBasis",
        "Definition",
        "PricingParameters",
        "RepoCurveType",
        "RepoParameters",
        "UnderlyingContract",
        "UnderlyingPricingParameters",
    ]
    testing_attributes = dir(rdf.repo)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, testing_attributes


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "universe": [
            {
                "instrumentDefinition": {
                    "startDate": "start_date",
                },
                "extended_params": "extended_params",
                "instrumentType": "Repo",
                "pricingParameters": {"marketDataDate": "market_data_date"},
            }
        ],
    }
    definition = rdf.repo.Definition(
        start_date="start_date",
        fields=["1", "2"],
        pricing_parameters=rdf.repo.PricingParameters(
            market_data_date="market_data_date"
        ),
        extended_params={"extended_params": "extended_params"},
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body



import pytest

from refinitiv.data.delivery._data._data_provider import ParsedData
from refinitiv.data.delivery._stream.stream_cache import StreamCache
from refinitiv.data.content.pricing._pricing_content_provider import (
    PricingRequestFactory,
    PricingContentValidator,
    PricingResponseFactory,
    PriceCache,
)


def test_pricing_request_factory():
    # given
    expected_value = [("universe", "sun,moon"), ("fields", "corn,rice")]
    factory = PricingRequestFactory()

    # when
    testing_value = factory.get_query_parameters(
        universe=["sun", "moon"], fields=["corn", "rice"]
    )

    # then
    assert testing_value == expected_value, testing_value


def test_pricing_content_validator_return_false():
    # given
    validator = PricingContentValidator()
    input_value = ParsedData("Error", {}, {})
    expected_value = False

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value, testing_value


def test_pricing_content_validator_add_error_message():
    # given
    validator = PricingContentValidator()
    code = 100500
    message = "invalid"
    input_value = ParsedData(
        "Error",
        {},
        {"status": "Error", "code": code, "message": message},
        code,
        message,
    )
    expected_value = {
        "status": "Error",
        "raw_response": {},
        "content_data": {"status": "Error", "code": 100500, "message": "invalid"},
        "error_codes": [100500],
        "error_messages": ["invalid"],
    }

    # when
    validator.validate(input_value)
    testing_value = input_value

    # then
    assert testing_value == expected_value, testing_value


def test_pricing_response_factory_create_not_empty_df():
    # given
    factory = PricingResponseFactory()
    input_value = ParsedData(
        {},
        {},
        [
            {
                "Fields": {"BID": 1.1},
                "Type": "Refresh",
                "State": {"Stream": "NonStreaming", "Data": "Ok"},
                "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            },
            {
                "Fields": {"BID": None},
                "Type": "Refresh",
                "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            },
            {
                "Fields": {"BID": 1.1},
                "Type": "Status",
                "State": {"Code": "NotEntitled"},
                "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            },
            {
                "Fields": {"BID": 1.1},
                "Type": "Status",
                "State": {"Code": "NotFound"},
                "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            },
            {
                "Fields": {"BID": 1.1},
                "Type": "__stub__",
                "Key": {"Service": "ELEKTRON_DD", "Name": "EUR="},
            },
        ],
    )

    # when
    testing_value = factory.create_success(
        input_value,
        universe=["sun", "moon", "mars", "jupiter", "earth"],
        fields=[],
    )

    # then
    assert not testing_value.data.df.empty


def test_price_cache_keys():
    cache = PriceCache({})

    keys = cache.keys()

    assert keys is not None


def test_price_cache_values():
    cache = PriceCache({})

    values = cache.values()

    assert values is not None


def test_price_cache_items():
    cache = PriceCache({})

    items = cache.items()

    assert items is not None


def test_price_cache___iter___raise_error():
    cache = PriceCache({})

    iterator = iter(cache)

    with pytest.raises(StopIteration):
        next(iterator)


def test_price_cache___iter___():
    expected_value = StreamCache("name")
    cache = PriceCache({"EUR=": expected_value})

    iterator = iter(cache)
    testing_value = next(iterator)

    assert testing_value == expected_value


def test_price_cache___getitem__raise_error():
    cache = PriceCache({})
    key = "key"

    with pytest.raises(KeyError, match=f"{key} not in PriceCache"):
        cache[key]


def test_price_cache___getitem__():
    expected_value = "value"
    cache = PriceCache({"key": expected_value})

    testing_value = cache["key"]

    assert testing_value == expected_value, testing_value


def test_price_cache___len__():
    cache = PriceCache({})

    length = len(cache)

    assert length == 0


def test_price_cache___str__():
    cache = PriceCache({})

    s = str(cache)

    assert s == "{}"



from ._historical_pricing_request_factory import (
    HistoricalPricingEventsRequestFactory,
    HistoricalPricingSummariesRequestFactory,
)
from .._historical_content_validator import HistoricalContentValidator
from .._historical_data_provider import (
    EventsDataProvider,
    SummariesDataProvider,
)
from .._historical_response_factory import HistoricalResponseFactory

hp_events_data_provider = EventsDataProvider(
    request=HistoricalPricingEventsRequestFactory(),
    response=HistoricalResponseFactory(),
    validator=HistoricalContentValidator(),
)

hp_summaries_data_provider = SummariesDataProvider(
    request=HistoricalPricingSummariesRequestFactory(),
    response=HistoricalResponseFactory(),
    validator=HistoricalContentValidator(),
)



class GlobalSettings(object):
    def __init__(self, *args, **kwargs):
        self._edp_user = kwargs.get("edp_user")
        self._edp_pwd = kwargs.get("edp_password")
        self._app_key = kwargs.get("app_key")
        self._trep_host = kwargs.get("trep_host")

    def edp_user(self):
        return self._edp_user

    def edp_password(self):
        return self._edp_pwd

    def app_key(self):
        return self._app_key

    def trep_host(self):
        return self._trep_host



from abc import ABC, abstractmethod
from typing import List

from refinitiv.data.usage_collection._utils import LogRecord


class AbstractUsageLogger(ABC):
    """
    Abstract class for usage loggers.
    Should be implemented by the user.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def log(self, batch: List[LogRecord]) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



# coding: utf8

from enum import Enum, unique


@unique
class FxCrossType(Enum):
    FX_FORWARD = "FxForward"
    FX_NON_DELIVERABLE_FORWARD = "FxNonDeliverableForward"
    FX_SPOT = "FxSpot"
    FX_SWAP = "FxSwap"
    FX_TIME_OPTION_FORWARD = "FxTimeOptionForward"



import refinitiv.data.content.ipa.financial_contracts as rdf


def term_deposit_definition():
    definition = rdf.term_deposit.Definition(
        instrument_tag="AED_AM1A",
        tenor="5Y",
        notional_ccy="GBP",
        fixed_rate_percent=1,
        pricing_parameters=rdf.term_deposit.PricingParameters(
            valuation_date="2018-01-10T00:00:00Z"
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "InterestAmountInDealCcy",
            "FixedRate",
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "ErrorMessage",
        ],
    )

    return definition


term_deposit_universe = {
    "instrumentType": "TermDeposit",
    "instrumentDefinition": {
        "instrumentTag": "my tag AED_AM1A",
        "tenor": "5Y",
        "notionalCcy": "GBP",
        "fixedRatePercent": 1,
    },
    "pricingParameters": {"valuationDate": "2018-01-10T00:00:00Z"},
}



import pytest
from http.server import HTTPServer, BaseHTTPRequestHandler
import pytest
from functools import partial
from threading import Thread

success_param = [
    ({"filename": "filename"}),
    ({"dcn": "dcn"}),
    ({"doc_id": "doc_id"}),
    ({"filing_id": "filing_id"}),
]

error_param = [
    ({}),
    (
        {
            "filename": "filename",
            "dcn": "dcn",
            "doc_id": "doc_id",
            "filing_id": "filing_id",
        }
    ),
    (
        {
            "filename": "filename",
            "dcn": "dcn",
        }
    ),
    (
        {
            "filename": "filename",
            "doc_id": "doc_id",
        }
    ),
    (
        {
            "filename": "filename",
            "filing_id": "filing_id",
        }
    ),
    (
        {
            "dcn": "dcn",
            "doc_id": "doc_id",
        }
    ),
    (
        {
            "dcn": "dcn",
            "filing_id": "filing_id",
        }
    ),
    (
        {
            "doc_id": "doc_id",
            "filing_id": "filing_id",
        }
    ),
    (
        {
            "filename": "filename",
            "dcn": "dcn",
            "doc_id": "doc_id",
        }
    ),
]


@pytest.fixture(params=success_param)
def create_filings_definition(request):
    try:
        yield request.param
    except Exception as e:
        assert False, str(e)


@pytest.fixture(params=error_param)
def create_filings_definition_with_error(request):
    return request.param


class HTTPHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/plain")
        self.end_headers()
        self.wfile.write("test text".encode("utf-8"))


class HTTPHandlerError(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(404)
        self.send_header("Content-type", "text/plain")
        self.end_headers()
        self.wfile.write("File not found".encode("utf-8"))


@pytest.fixture(scope="module")
def mock_server_success(handler=HTTPHandler):
    # Passing 0 to server port argument claims a random free port
    server = HTTPServer(("127.0.0.1", 0), handler)
    server_thread = Thread(
        target=partial(server.serve_forever, poll_interval=0.1), daemon=True
    )
    server_thread.start()
    yield server
    server.shutdown()


@pytest.fixture(scope="module")
def mock_server_error(handler=HTTPHandlerError):
    # Passing 0 to server port argument claims a random free port
    server = HTTPServer(("127.0.0.1", 0), handler)
    server_thread = Thread(
        target=partial(server.serve_forever, poll_interval=0.1), daemon=True
    )
    server_thread.start()
    yield server
    server.shutdown()



from ._session_cxn_type import SessionCxnType
from ._session_type import SessionType
from ._session_definition import Definition

from ._dacs_params import DacsParams
from ._session import Session
from .event import UpdateEvent
from . import auth_manager
from .tools import is_open, is_closed, raise_if_closed
from ._retry_transport import RetryTransport, RetryAsyncTransport
from ._desktop_session import DesktopSession
from ._platform_session import PlatformSession

from .grant_password import *

from ._default_session_manager import (
    get_default,
    set_default,
    _eikon_default_session_manager,
    _rd_default_session_manager,
    get_valid_session,
    EikonDefaultSessionManager,
    RDDefaultSessionManager,
)

from .connection import *



from typing import TYPE_CHECKING

from .._definition import BaseSurfaceDefinition
from ..._surfaces._cap_surface_request_item import CapSurfaceRequestItem
from ....._content_type import ContentType

if TYPE_CHECKING:
    from ..._surfaces._surface_types import SurfaceParameters, SurfaceLayout
    from ....._types import ExtendedParams, OptStr


class Definition(BaseSurfaceDefinition):
    def __init__(
        self,
        surface_layout: "SurfaceLayout" = None,
        surface_parameters: "SurfaceParameters" = None,
        underlying_definition: dict = None,
        surface_tag: "OptStr" = None,
        instrument_type=None,
        extended_params: "ExtendedParams" = None,
    ):
        request_item = CapSurfaceRequestItem(
            instrument_type=instrument_type,
            surface_layout=surface_layout,
            surface_params=surface_parameters,
            underlying_definition=underlying_definition,
            surface_tag=surface_tag,
        )
        super().__init__(
            content_type=ContentType.SURFACES,
            universe=request_item,
            extended_params=extended_params,
        )



import json
import os
from importlib import import_module

import pytest

from refinitiv.data import _configure as configure

RDPLIB_ENV = "RDPLIB_ENV"
RDPLIB_ENV_DIR = "RDPLIB_ENV_DIR"
RD_LIB_CONFIG_PATH = "RD_LIB_CONFIG_PATH"

expected_default_config_filename = "refinitiv-data.config.json"
expected_default_user_config_path = os.path.join(
    os.path.expanduser("~"), expected_default_config_filename
)
expected_default_project_config_path = os.path.join(
    os.getcwd(), expected_default_config_filename
)


def annotate(**kwargs):
    return list(kwargs.values())


@pytest.fixture(autouse=True)
def setup_teardown_function(monkeypatch):
    monkeypatch.delenv(RDPLIB_ENV, raising=False)
    monkeypatch.delenv(RDPLIB_ENV_DIR, raising=False)
    monkeypatch.delenv(RD_LIB_CONFIG_PATH, raising=False)

    configure.reload()

    yield

    remove_user_config()
    remove_project_config()


def get_client_config_path(configure):
    return configure._config_files_paths[0]


def get_project_config_path(configure):
    return configure._config_files_paths[0]


def get_user_config_path(configure):
    return configure._config_files_paths[1]


def get_default_config_path(configure):
    return configure._config_files_paths[2]


def remove_config(path):
    while os.path.exists(path):
        try:
            os.remove(path)
        except (PermissionError, FileNotFoundError):
            pass


def remove_user_config():
    remove_config(expected_default_user_config_path)


def remove_project_config():
    remove_config(expected_default_project_config_path)


def write_config(path, s):
    if isinstance(s, dict):
        s = json.dumps(s)
    with open(path, "w") as f:
        f.write(s)

    return write_config


def write_user_config(s):
    write_config(expected_default_user_config_path, s)


def write_project_config(s):
    write_config(expected_default_project_config_path, s)


ENV_NAME = "USER"
ENV_VALUE = "TestingUser"


@pytest.fixture
def mock_env_user(monkeypatch):
    monkeypatch.setenv(ENV_NAME, ENV_VALUE)


@pytest.fixture
def mock_env_missing(monkeypatch):
    monkeypatch.delenv(ENV_NAME, raising=False)



__all__ = (
    "Definition",
    "CreditCurveDefinition",
    "CreditCurveParameters",
    "BasisSplineSmoothModel",
    "BusinessSector",
    "ExtrapolationMode",
    "CalendarAdjustment",
    "CalibrationModel",
    "CompoundingType",
    "CurveSubType",
    "EconomicSector",
    "Industry",
    "IndustryGroup",
    "InterestCalculationMethod",
    "InterpolationMode",
    "IssuerType",
    "MainConstituentAssetClass",
    "PriceSide",
    "Rating",
    "RatingScaleSource",
    "ReferenceEntityType",
    "Seniority",
)

from ._definition import Definition
from ...._curves._bond_curves import (
    CreditCurveDefinition,
    CreditCurveParameters,
)

from ._enums import (
    BasisSplineSmoothModel,
    BusinessSector,
    ExtrapolationMode,
    CalendarAdjustment,
    CalibrationModel,
    CompoundingType,
    CurveSubType,
    EconomicSector,
    Industry,
    IndustryGroup,
    InterestCalculationMethod,
    InterpolationMode,
    IssuerType,
    MainConstituentAssetClass,
    PriceSide,
    Rating,
    RatingScaleSource,
    ReferenceEntityType,
    Seniority,
)



"""Module to create and use templated search presets for the discovery search

Use templated search from the "Equity" template that was defined in the config:
>>> import refinitiv.data as rd
>>> rd.discovery.search_templates["Equity"].search()
"""

from refinitiv.data.discovery._search_templates.manage import SearchTemplates

__all__ = ["templates"]

templates = SearchTemplates()



import pytest

from refinitiv.data.discovery._search_templates.namespaces import Namespace
from refinitiv.data.discovery._search_templates.search import DiscoverySearchTemplate


class TestBasicNameSpace:
    def test_basic(self):
        ns = Namespace(
            locals=Namespace(
                One=DiscoverySearchTemplate(name="One"),
            ),
            user=Namespace(
                sub=Namespace(
                    Two=DiscoverySearchTemplate(name="Two"),
                ),
                Three=DiscoverySearchTemplate(name="Three"),
            ),
        )
        assert ns.get("user.sub.Two")
        assert not ns.get("user.sub.Two.val")
        assert set(ns) == {"locals", "user"}

    def test_keys_must_be_strings(self):
        with pytest.raises(TypeError):
            Namespace(**{1: DiscoverySearchTemplate("1")})

    def test_names_cant_contain_dots(self):
        with pytest.raises(ValueError):
            Namespace(**{"a.b": DiscoverySearchTemplate("1")})

    def test_cant_have_arbitrary_types(self):
        with pytest.raises(TypeError):
            Namespace(abc=123)



__all__ = (
    "AbstractUsageLogger",
    "FilterType",
    "LoggerConfig",
    "LogRecord",
    "RecordData",
    "get_usage_logger",
    "StreamUsageKey",
)

from ._abstract_logger import AbstractUsageLogger
from ._filter_types import FilterType
from ._utils import LogRecord, LoggerConfig, RecordData, StreamUsageKey
from ._logger import get_usage_logger



from typing import TYPE_CHECKING

from .._arg_enums import main_constituent_asset_class_arg_parser, risk_type_arg_parser
from .._base_data_class import BaseData
from ......delivery._data._endpoint_data import EndpointData

if TYPE_CHECKING:
    from ...._curves._cross_currency_curves._types import (
        OptMainConstituentAssetClass,
        OptRiskType,
    )
    from ......_types import OptStr, OptBool

_camel_to_snake = {
    "mainConstituentAssetClass": "main_constituent_asset_class",
    "riskType": "risk_type",
    "baseCurrency": "base_currency",
    "baseIndexName": "base_index_name",
    "id": "id",
    "isFallbackForFxCurveDefinition": "is_fallback_for_fx_curve_definition",
    "isNonDeliverable": "is_non_deliverable",
    "name": "name",
    "quotedCurrency": "quoted_currency",
    "quotedIndexName": "quoted_index_name",
    "source": "source",
}


def convert_camel_to_snake(data: dict) -> dict:
    return {_camel_to_snake.get(name, name): value for name, value in data.items()}


class CurveDefinitionTriangulate(BaseData):
    """
    Gets the Cross Currency curve definitions who can be used with provided criterias.

    Parameters
    ----------
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    id : str, optional
        The identifier of the cross currency definitions.
    is_fallback_for_fx_curve_definition : bool, optional

    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    """

    def __init__(
        self,
        main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
        risk_type: "OptRiskType" = None,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        id: "OptStr" = None,
        is_fallback_for_fx_curve_definition: "OptBool" = None,
        is_non_deliverable: "OptBool" = None,
        name: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        source: "OptStr" = None,
        **kwargs,
    ):
        self.main_constituent_asset_class = (
            main_constituent_asset_class_arg_parser.get_enum(
                main_constituent_asset_class
            )
        )
        self.risk_type = risk_type_arg_parser.get_enum(risk_type)
        self.base_currency = base_currency
        self.base_index_name = base_index_name
        self.id = id
        self.is_fallback_for_fx_curve_definition = is_fallback_for_fx_curve_definition
        self.is_non_deliverable = is_non_deliverable
        self.name = name
        self.quoted_currency = quoted_currency
        self.quoted_index_name = quoted_index_name
        self.source = source
        super().__init__(**kwargs)


class CurveDefinition:
    def __init__(self, **kwargs):
        self._direct_curve_definitions = None
        self._indirect_curve_definitions = None
        self._kwargs = kwargs

    @property
    def direct_curve_definitions(self):
        if self._direct_curve_definitions is None:
            self._direct_curve_definitions = []
            direct_curves = self._kwargs.get("directCurveDefinitions")
            self._direct_curve_definitions = self._create_list_definition_triangulates(
                direct_curves
            )
        return self._direct_curve_definitions

    @property
    def indirect_curve_definitions(self):
        if self._indirect_curve_definitions is None:
            self._indirect_curve_definitions = []
            indirect_curves = self._kwargs.get("indirectCurveDefinitions")
            for indirect_curve in indirect_curves:
                cross_currencies = indirect_curve.get("crossCurrencyDefinitions", [])
                list_triangulates = self._create_list_definition_triangulates(
                    cross_currencies
                )
                self._indirect_curve_definitions.append(list_triangulates)
        return self._indirect_curve_definitions

    @property
    def curve_tag(self):
        return self._kwargs.get("curveTag")

    def _create_list_definition_triangulates(self, items: list):
        return [
            CurveDefinitionTriangulate(**convert_camel_to_snake(item)) for item in items
        ]


class TriangulateDefinitionsData(EndpointData):
    _curve_definitions = None

    @property
    def curve_definitions(self):
        if self._curve_definitions is None:
            self._curve_definitions = [
                CurveDefinition(**item) for item in self._raw.get("data", [])
            ]
        return self._curve_definitions



from unittest.mock import Mock, patch

import pytest

from refinitiv.data.delivery._data._data_factory import DataFactory
from refinitiv.data.delivery._data._data_provider import (
    Response,
    RequestFactory,
    ParsedData,
)
from refinitiv.data.delivery._data._endpoint_data import RequestMethod
from tests.unit.conftest import StubSession, args


@pytest.mark.parametrize(
    "arg",
    [
        "data",
        "is_success",
        "errors",
        "requests_count",
        "request_message",
        "closure",
        "http_status",
        "http_headers",
    ],
)
def test_response_args(arg):
    # given
    # when
    response = Response(True, ParsedData({}, {}), DataFactory())

    # then
    assert hasattr(response, arg)


def test_response_requests_count_not_list():
    # given
    # when
    response = Response(True, ParsedData({}, {}), DataFactory())

    # then
    assert response.requests_count == 1


def test_response_requests():
    # given
    count = 10
    # when
    response = Response(True, ParsedData({}, {}), DataFactory())
    response.http_response = list(range(count))

    # then
    assert response.requests_count == count


def test_extend_query_parameters_method_for_request_factory():
    class StubRequestFactory(RequestFactory):
        def get_query_parameters(self, *args, **kwargs):
            return kwargs.get("query_parameters") or {}

    # given
    session = StubSession(is_open=True)
    request_factory = StubRequestFactory()
    request_factory.extend_query_parameters = Mock()
    request_factory.update_url = Mock()

    query_parameters = ("parameter", "value")
    extended_params = {"test": "value"}
    kwargs = {"query_parameters": query_parameters, "extended_params": extended_params}

    # when
    request_factory.create(session, ..., **kwargs)

    # then
    request_factory.extend_query_parameters.assert_called_once_with(
        query_parameters, extended_params=extended_params
    )


@patch("refinitiv.data.delivery._data._request_factory.Request")
def test_body_parameters_method_create_for_request_factory_default(patch_request_class):
    # given
    patch_request_class.return_value = Mock()

    session = StubSession(is_open=True)
    request_factory = RequestFactory()
    request_factory.update_url = Mock(return_value="url")

    # when
    request_factory.create(session, ..., **{})

    # then
    patch_request_class.assert_called_once_with(
        url=request_factory.update_url.return_value,
        method=RequestMethod.GET,
        headers={},
        json=None,
    )


@pytest.mark.parametrize(
    "method,headers,body_parameters",
    [
        args(
            method=RequestMethod.GET,
            headers={},
            body_parameters=None,
        ),
        args(
            method=RequestMethod.POST,
            headers={"Content-Type": "application/json"},
            body_parameters={},
        ),
        args(
            method=RequestMethod.PUT,
            headers={"Content-Type": "application/json"},
            body_parameters={},
        ),
        args(
            method=RequestMethod.DELETE,
            headers={"Content-Type": "application/json"},
            body_parameters=None,
        ),
    ],
)
@patch("refinitiv.data.delivery._data._request_factory.Request")
def test_body_parameters_method_create_for_request_factory(
    patch_request_class, method, headers, body_parameters
):
    # given
    patch_request_class.return_value = Mock()

    session = StubSession(is_open=True)
    request_factory = RequestFactory()
    request_factory.update_url = Mock(return_value="url")
    kwargs = {"method": method}

    # when
    request_factory.create(session, ..., **kwargs)

    # then
    patch_request_class.assert_called_once_with(
        url=request_factory.update_url.return_value,
        method=method,
        headers=headers,
        json=body_parameters,
    )



import math
import numpy as np

import refinitiv.data as rd
from refinitiv.data import OpenState
from refinitiv.data._fin_coder_layer._intervals_consts import NON_INTRA_DAY_INTERVALS
from tests.integration.helpers import compare_list


OHLC_COLUMNS_NAMES = ["close", "high", "low", "open"]


def check_session_state(session):
    assert session.open_state == rd.OpenState.Opened, f"Session is not open"

    rd.close_session()
    assert session.open_state == rd.OpenState.Closed, f"Session is open"


def check_column_names_is_exist_in_response_and_df_not_empty(
    response, expected_fields=None
):
    assert response is not None
    assert not response.empty, f"DataFrame is empty: {response}"

    if expected_fields != [] and expected_fields is not None:
        if isinstance(expected_fields, str):
            expected_fields = [expected_fields]
        actual_fields = [col for col in response.columns]
        compare_list(list(set(actual_fields)), expected_fields)


def check_rics_fields_order_respects_and_df_not_empty(
    response, expected_universes, expected_fields
):
    actual_rics, actual_fields = [], []
    for ric, field in response.columns:
        if ric not in actual_rics:
            actual_rics.append(ric)
        if field not in actual_fields:
            actual_fields.append(field)

    assert response is not None
    assert not response.empty, f"DataFrame is empty: {response}"
    if not expected_universes == [] or not None:
        assert actual_rics == expected_universes, f"Inconsistency in the ric's order"
    if not expected_fields == [] or not None:
        assert actual_fields == expected_fields, f"Inconsistency in the field's order"


def check_stream_opened(stream):
    assert (
        stream._stream.open_state == OpenState.Opened
    ), "Stream is not in opened state."


def check_stream_closed(stream):
    assert (
        stream._stream.open_state == OpenState.Closed
    ), "Stream is not in closed state."


def check_stream_universe(stream, expected_universe):
    if isinstance(expected_universe, str):
        expected_universe = [expected_universe]

    universe_streams = stream._stream._stream._stream_by_name
    for universe in expected_universe:
        assert universe in universe_streams, f"Universe {universe} is not in stream"
        content = universe_streams[universe]
        assert content is not None, f"Data for universe {universe} is None"


def check_stream_fields(stream, universe, expected_fields):
    if isinstance(universe, str):
        universe = [universe]
    snapshot_df = stream.get_snapshot(universe, expected_fields)
    actual_fields = set(snapshot_df.columns)
    actual_fields.remove("Instrument")

    if isinstance(expected_fields, str):
        expected_fields = [expected_fields]

    if expected_fields:
        assert actual_fields == set(
            expected_fields
        ), f"{expected_fields} != {actual_fields}"

    else:
        assert actual_fields, f"{actual_fields} is empty list"


def check_stream_snapshot(stream, requested_universe, requested_fields):
    df = stream.get_snapshot(requested_universe, requested_fields)
    actual_universe = list(df["Instrument"])
    assert all(universe in actual_universe for universe in requested_universe)
    assert all(field in df for field in requested_fields)


def is_universe_exist(dataframe, expected_universe):
    universe_list = [
        column[0]
        for column in dataframe.columns.tolist()
        if "ticks count" not in column
    ]
    return list(set(universe_list)) == expected_universe


def check_df_index_names(response, interval=None):
    index_date = response.index

    if interval is not None and interval in NON_INTRA_DAY_INTERVALS:
        assert (
            index_date.name == "Date" or "date"
        ), f"Index column has the {index_date.name} name"
    else:
        assert (
            index_date.name == "Timestamp"
        ), f"Index column has the {index_date.name} name"


def check_df_recording_duration(elapsed_time, expected_duration):
    if isinstance(expected_duration, str):
        duration = expected_duration.rstrip("s")
        expected_duration = int(duration)
    assert (
        round(elapsed_time) == expected_duration
    ), f"Duration parameter {expected_duration} is not equal to actual elapsed time {elapsed_time}"


def check_df_recording_frequency(dataframe, expected_frequency):
    seconds_duration = dataframe.index.floor("S")
    difference = seconds_duration.to_series().diff()[1:]
    assert np.all(
        difference[0] == difference
    ), f"Frequency is not equal between all updates"
    assert difference[0].seconds == int(
        expected_frequency.rstrip("s")
    ), f"Actual frequency {difference[0].seconds} seconds is not equal to expected"


def check_history_for_expected_amount_of_bars(
    stream, history, frequency, duration, recorders_amount=1
):
    assert (
        history is stream.recorder._ohlc_builder.ohlc_df
    ), "There is inconsistency in recorded data with history"
    if isinstance(duration, str):
        duration = duration.rstrip("s")
        duration = int(duration)
    frequency = frequency.rstrip("s")
    frequency = int(frequency)
    count = math.ceil(duration / frequency)
    expected_amount = count * recorders_amount
    actual_amount = len(history)
    assert (
        actual_amount == expected_amount
    ), f"There is inconsistency with amount of ohlc bars, actual_amount={actual_amount}"


def check_triggered_events(callback_history, expected_universes):
    assert callback_history is not None, f"List of callbacks is empty"

    for df in callback_history:
        assert df is not None, f"Empty df is received"
        assert not df.empty, f"Empty df is received"

        universe = df.axes[0].values[0]
        assert universe in expected_universes, f"Universe is not expected in DataFrame"


def check_df_contains_ohlc_columns_and_universes(
    dataframe, expected_fields, expected_universe
):
    universe, fields, column_names = dataframe.axes[1].levels
    compare_list(column_names, OHLC_COLUMNS_NAMES)
    compare_list(fields, expected_fields)
    compare_list(universe, expected_universe)



from types import SimpleNamespace

import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._df_build_type import DFBuildType
from refinitiv.data.content._universe_stream import _UniverseStream
from refinitiv.data.content._universe_streams import _UniverseStreams
from tests.unit.conftest import is_dunder_method, StubSession


@pytest.fixture(
    scope="function",
    params=[
        ContentType.FORWARD_CURVE,
        ContentType.ZC_CURVES,
        ContentType.ZC_CURVE_DEFINITIONS,
        ContentType.SURFACES,
        ContentType.CONTRACTS,
    ],
)
def content_type(request):
    return request.param


@pytest.fixture(
    scope="function",
    params=[
        DFBuildType.INDEX,
        DFBuildType.DATE_AS_INDEX,
    ],
)
def dfbuild_type(request):
    return request.param


@pytest.fixture(scope="function")
def session_mock():
    from refinitiv.data import _configure as configure

    session_mock_ = StubSession(config=configure.get_config())
    return session_mock_


def remove_dunder_methods(iterable):
    return [item for item in iterable if not is_dunder_method(item)]


def remove_private_attributes(iterable):
    return [item for item in iterable if not item.startswith("_")]


@pytest.fixture(scope="function")
def stream():
    session = StubSession(is_open=True, fail_if_error=True)
    streaming_price = _UniverseStream(
        content_type=ContentType.NONE, name="name", session=session
    )
    return streaming_price


@pytest.fixture(scope="function")
def streams():
    session = StubSession(is_open=True, fail_if_error=True)
    _stream = _UniverseStreams(
        content_type=ContentType.NONE, universe="name", session=session
    )
    return _stream


TIMEOUT = 0.0001


class StubDefinition:
    def __init__(self) -> None:
        ns = SimpleNamespace()
        ns._get_enum_parameter = lambda *args: []
        self.surface_parameters = ns

    def get_dict(self):
        return {}

    def get_instrument_type(self):
        return ""



import abc
import threading
from typing import Union


class Updater(abc.ABC):
    def __init__(self, delay: Union[int, float], name: str = None):
        self.delay: Union[int, float] = delay or 1
        self.name = name or "updater"
        self._stopped: bool = True
        self._disposed: bool = False
        self._timeout_evt: threading.Event = threading.Event()

    @property
    def is_started(self) -> bool:
        return not self._stopped and not self._disposed

    @property
    def delay(self) -> Union[int, float]:
        return self._delay

    @delay.setter
    def delay(self, value: Union[int, float]):
        self._delay = value

    def start(self) -> None:
        if self._disposed:
            raise RuntimeError(f"{self.name} disposed")

        if self.is_started:
            return

        self._stopped = False
        self._timeout_evt.clear()

        while not self._stopped:
            self._timeout_evt.wait(timeout=self.delay)

            if not self._stopped:
                try:
                    self._do_update()
                except Exception as e:
                    self.stop()
                    raise e

    def stop(self) -> None:
        if self._disposed:
            raise RuntimeError(f"{self.name} disposed")

        if self._stopped:
            return

        self._stopped = True
        self._timeout_evt.set()

    def dispose(self) -> None:
        if self._disposed:
            return

        if not self._stopped:
            self.stop()

        self._disposed = True
        self._do_dispose()
        self._timeout_evt = None

    @abc.abstractmethod
    def _do_update(self) -> None:
        # for override
        pass

    @abc.abstractmethod
    def _do_dispose(self) -> None:
        # for override
        pass


class UpdaterThreaded(abc.ABC):
    def __init__(self, delay: Union[int, float], name: str = None):
        self.delay: Union[int, float] = delay or 1
        self.name = name or "updater"
        self._stopped: bool = True
        self._disposed: bool = False
        self._timeout_evt: threading.Event = threading.Event()
        self._start_evt: threading.Event = threading.Event()
        self._thread: threading.Thread = threading.Thread(
            target=self._run,
            name=f"{self.name}-Thread",
            daemon=True,
        )

    @property
    def delay(self) -> Union[int, float]:
        return self._delay

    @delay.setter
    def delay(self, value: Union[int, float]):
        self._delay = value

    def start(self) -> None:
        if self._disposed:
            raise RuntimeError(f"{self.name} disposed")

        if self._start_evt.is_set():
            return

        self._stopped = False
        self._timeout_evt.clear()
        self._start_evt.set()

        if not self._thread.ident:
            self._thread.start()

    def _run(self):
        while not self._disposed:
            self._start_evt.wait()

            while not self._stopped:
                self._timeout_evt.wait(timeout=self.delay)

                if not self._stopped:
                    try:
                        self._do_update()
                    except Exception as e:
                        self.stop()
                        raise e

    def stop(self) -> None:
        if self._disposed:
            raise RuntimeError(f"{self.name} disposed")

        if self._stopped:
            return

        self._stopped = True
        self._timeout_evt.set()
        self._start_evt.clear()

    def dispose(self) -> None:
        if self._disposed:
            return

        if not self._stopped:
            self.stop()

        self._disposed = True
        self._do_dispose()
        self._thread = None
        self._start_evt = None
        self._timeout_evt = None

    @abc.abstractmethod
    def _do_update(self) -> None:
        # for override
        pass

    @abc.abstractmethod
    def _do_dispose(self) -> None:
        # for override
        pass



from datetime import datetime

import pandas as pd
import pytest

from refinitiv.data import get_config
from refinitiv.data.delivery._data._data_provider import Response
from tests.integration.helpers import compare_list


def get_column_names_from_list(data_frame):
    df = pd.DataFrame(data_frame)
    df_list = []
    for col in df.columns:
        if col != "Instrument":
            df_list.append(col)
    return df_list


def check_df_column_contains_value(df, column_name, expected_value):
    column_values = df[column_name].tolist()

    if isinstance(expected_value, list):
        assert column_values == expected_value, f"{column_values} != {expected_value}"

    if isinstance(expected_value, str):
        for item in column_values:
            assert item == expected_value, f"{item} != {expected_value}"


def check_pricing_data(pricing_definition, expected_universe, expected_fields=None):
    df = None

    if isinstance(pricing_definition, Response):
        df = pricing_definition.data.df

    elif isinstance(pricing_definition, pd.DataFrame):
        df = pricing_definition

    actual_fields = get_column_names_from_list(df)

    if expected_fields is None:
        assert not df.empty

    elif isinstance(expected_fields, str):
        actual_fields = actual_fields[0]
        assert expected_fields == actual_fields, f"{expected_fields} != {actual_fields}"

    elif isinstance(expected_fields, list):
        compare_list(actual_fields, expected_fields)

    check_df_column_contains_value(df, "Instrument", expected_universe)


def check_item_message_and_status(universe, expected_message, pricing_stream):
    actual_message = None
    actual_item_status = None

    if isinstance(universe, str):
        actual_message = pricing_stream._stream._stream_by_name[
            universe
        ].message.replace("**", "")
        actual_item_status = str(pricing_stream._stream._stream_by_name[universe].state)

    if isinstance(universe, list):
        actual_message = pricing_stream._stream._stream_by_name[
            universe[1]
        ].message.replace("**", "")
        actual_item_status = str(
            pricing_stream._stream._stream_by_name[universe[1]].state
        )

    assert actual_message == expected_message, f"{actual_message}"
    assert (
        actual_item_status == "StreamState.Opened"
    ), f"Stream state is not closed, {actual_item_status}"


def on_event(msg, instrument_name, stream, triggered_events_list=None):
    current_time = datetime.now().time()
    if triggered_events_list is not None:
        triggered_events_list.append(msg)
    print(
        f"{current_time} - Stream {stream} - Received event : {msg} for RIC: {instrument_name}"
    )


def add_callback_for_contribution(stream, event_list):
    stream.on_ack(
        lambda status, instrument_name, stream_item: on_event(
            status, instrument_name, stream_item, event_list
        )
    )
    stream.on_error(
        lambda status, instrument_name, stream_item: on_event(
            status, instrument_name, stream_item, event_list
        )
    )

    return stream


@pytest.fixture(scope="function")
def setup_direct_url():
    config = get_config()
    config.set_param(
        "apis.streaming.pricing.endpoints.main",
        {"direct-url": "ws://10.184.8.188:15000"},
        auto_create=True,
    )



# coding: utf8

__all__ = ["DateScheduleFrequency"]

from enum import Enum, unique


@unique
class DateScheduleFrequency(Enum):
    """
    The frequency of dates in the predefined period. The possible values are:

    The possible values are:
        - Daily,
        - Weekly,
        - BiWeekly,
        - Monthly,
    """

    DAILY = "Daily"
    WEEKLY = "Weekly"
    BI_WEEKLY = "BiWeekly"
    MONTHLY = "Monthly"



import pandas as pd
import pytest

import refinitiv.data.content as rdc
from refinitiv.data._core.session import SessionType
from refinitiv.data.content import estimates
from refinitiv.data.content import ownership
from refinitiv.data.content import search

from tests.unit.conftest import StubSession, args, StubResponse, StubConfig
from tests.unit.content.data_for_tests import (
    NEWS_STORY_RESPONSE,
    FUNDAMENTAL_AND_REFERENCE_DEFINITION,
    NEWS_HEADLINES_DEFINITION,
    NEWS_HEADLINES_RESPONSE,
    NEWS_STORY_DEFINITION,
    SURFACES_DEFINITION,
    SURFACE_SWAPTION_DEFINITION,
    SURFACE_SWAPTION_RESPONSE,
    SURFACES_RESPONSE,
    PRICING_DEFINITION,
    PRICING_RESPONSE,
    FORWARD_CURVE_RESPONSE,
    FORWARD_CURVES_RESPONSE,
    ZC_CURVE_RESPONSE,
    ZC_CURVE_DEFINITION_RESPONSE,
    ZC_CURVE_DEFINITIONS_RESPONSE,
    ZC_CURVES_RESPONSE,
    SURFACE_CAP_DEFINITION,
    SURFACE_CAP_RESPONSE,
    SURFACE_ETI_DEFINITION,
    SURFACE_ETI_RESPONSE,
    SURFACE_FX_DEFINITION,
    SURFACE_FX_RESPONSE,
    NEWS_HEADLINES_RESPONSE_UDF,
    FUNDAMENTAL_AND_REFERENCE_RESPONSE_DESKTOP,
    FUNDAMENTAL_AND_REFERENCE_RESPONSE_PLATFORM,
)


def assert_create_df_on_demand(response):
    assert response.data._dataframe is None

    # when
    df = response.data.df

    # then
    assert df is not None
    assert isinstance(response.data._dataframe, pd.DataFrame)


def assert_create_df_immediately(response):
    assert response.data._dataframe is not None

    # when
    df = response.data.df

    # then
    assert df is not None
    assert isinstance(response.data._dataframe, pd.DataFrame)


def assert_create_empty_df(response):
    assert response.data._dataframe is None

    # when
    df = response.data.df

    # then
    assert df is not None
    assert isinstance(response.data._dataframe, pd.DataFrame)
    assert df.empty


@pytest.mark.parametrize(
    argnames="definition",
    ids=[
        "custom_instruments.search.Definition",
        "esg.basic_overview.Definition",
        "esg.full_measures.Definition",
        "esg.full_scores.Definition",
        "esg.standard_measures.Definition",
        "esg.standard_scores.Definition",
        "esg.universe.Definition",
        "estimates.view_actuals.annual.Definition",
        "estimates.view_actuals.interim.Definition",
        "estimates.view_actuals_kpi.annual.Definition",
        "estimates.view_actuals_kpi.interim.Definition",
        "estimates.view_summary.annual.Definition",
        "estimates.view_summary.historical_snapshots_non_periodic_measures.Definition",
        "estimates.view_summary.historical_snapshots_periodic_measures_annual.Definition",
        "estimates.view_summary.historical_snapshots_periodic_measures_interim.Definition",
        "estimates.view_summary.historical_snapshots_recommendations.Definition",
        "estimates.view_summary.interim.Definition",
        "estimates.view_summary.non_periodic_measures.Definition",
        "estimates.view_summary.recommendations.Definition",
        "estimates.view_summary_kpi.annual.Definition",
        "estimates.view_summary_kpi.historical_snapshots_kpi.Definition",
        "estimates.view_summary_kpi.interim.Definition",
        "ownership.consolidated.breakdown.Definition",
        "ownership.consolidated.concentration.Definition",
        "ownership.consolidated.investors.Definition",
        "ownership.consolidated.recent_activity.Definition",
        "ownership.consolidated.shareholders_history_report.Definition",
        "ownership.consolidated.shareholders_report.Definition",
        "ownership.consolidated.top_n_concentration.Definition",
        "ownership.fund.breakdown.Definition",
        "ownership.fund.concentration.Definition",
        "ownership.fund.holdings.Definition",
        "ownership.fund.investors.Definition",
        "ownership.fund.recent_activity.Definition",
        "ownership.fund.shareholders_history_report.Definition",
        "ownership.fund.shareholders_report.Definition",
        "ownership.fund.top_n_concentration.Definition",
        "ownership.insider.shareholders_report.Definition",
        "ownership.insider.transaction_report.Definition",
        "ownership.investor.holdings.Definition",
        "ownership.org_info.Definition",
        "search.Definition",
        "search.lookup.Definition",
        "search.metadata.Definition",
        "symbol_conversion.Definition",
    ],
    argvalues=[
        # <editor-fold desc="custom_instruments">
        rdc.custom_instruments.search.Definition("VOD.L"),
        # </editor-fold>
        # <editor-fold desc="esg">
        rdc.esg.basic_overview.Definition("IBM.N"),
        rdc.esg.full_measures.Definition("BNPP.PA"),
        rdc.esg.full_scores.Definition(universe="4295904307", start=0, end=-5),
        rdc.esg.standard_measures.Definition("BNPP.PA"),
        rdc.esg.standard_scores.Definition("6758.T"),
        rdc.esg.universe.Definition(),
        # </editor-fold>
        # <editor-fold desc="estimates">
        rdc.estimates.view_actuals.annual.Definition("IBM.N", estimates.Package.BASIC),
        rdc.estimates.view_actuals.interim.Definition("IBM.N", estimates.Package.BASIC),
        rdc.estimates.view_actuals_kpi.annual.Definition("BNPP.PA"),
        rdc.estimates.view_actuals_kpi.interim.Definition("BNPP.PA"),
        rdc.estimates.view_summary.annual.Definition("IBM.N", estimates.Package.BASIC),
        rdc.estimates.view_summary.historical_snapshots_non_periodic_measures.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary.historical_snapshots_periodic_measures_annual.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary.historical_snapshots_periodic_measures_interim.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary.historical_snapshots_recommendations.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary.interim.Definition("IBM.N", estimates.Package.BASIC),
        rdc.estimates.view_summary.non_periodic_measures.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary.recommendations.Definition(
            "IBM.N", estimates.Package.BASIC
        ),
        rdc.estimates.view_summary_kpi.annual.Definition("ORCL.N"),
        rdc.estimates.view_summary_kpi.historical_snapshots_kpi.Definition("BNPP.PA"),
        rdc.estimates.view_summary_kpi.interim.Definition("BNPP.PA"),
        # </editor-fold>
        # <editor-fold desc="ownership">
        rdc.ownership.consolidated.breakdown.Definition(
            "TRI.N", ownership.StatTypes.INVESTOR_TYPE
        ),
        rdc.ownership.consolidated.concentration.Definition("TRI.N"),
        rdc.ownership.consolidated.investors.Definition("TRI.N"),
        rdc.ownership.consolidated.recent_activity.Definition("TRI.N", "asc"),
        rdc.ownership.consolidated.shareholders_history_report.Definition(
            "TRI.N", ownership.Frequency.MONTHLY
        ),
        rdc.ownership.consolidated.shareholders_report.Definition("TRI.N"),
        rdc.ownership.consolidated.top_n_concentration.Definition("TRI.N", 30),
        rdc.ownership.fund.breakdown.Definition(
            "TRI.N", ownership.StatTypes.INVESTOR_TYPE
        ),
        rdc.ownership.fund.concentration.Definition("TRI.N"),
        rdc.ownership.fund.holdings.Definition("LP40189339"),
        rdc.ownership.fund.investors.Definition("TRI.N"),
        rdc.ownership.fund.recent_activity.Definition(
            "TRI.N", ownership.SortOrder.ASCENDING
        ),
        rdc.ownership.fund.shareholders_history_report.Definition(
            "TRI.N", "M", start="-1Q"
        ),
        rdc.ownership.fund.shareholders_report.Definition("TRI.N"),
        rdc.ownership.fund.top_n_concentration.Definition("TRI.N", 30),
        rdc.ownership.insider.shareholders_report.Definition("TRI.N"),
        rdc.ownership.insider.transaction_report.Definition("TRI.N", start="-1Q"),
        rdc.ownership.investor.holdings.Definition("TRI.N"),
        rdc.ownership.org_info.Definition("TRI.N"),
        # </editor-fold>
        # <editor-fold desc="search">
        rdc.search.Definition(query="cfo", view=search.Views.PEOPLE),
        rdc.search.lookup.Definition(
            view=search.Views.SEARCH_ALL,
            scope="RIC",
            terms="A,B,NOSUCHRIC,C,D",
            select="BusinessEntity,DocumentTitle",
        ),
        rdc.search.metadata.Definition(search.Views.PEOPLE),
        # </editor-fold>
        # <editor-fold desc="symbol_conversion">
        rdc.symbol_conversion.Definition(["US5949181045", "US02079K1079"]),
        # </editor-fold>
    ],
)
def test_creating_df_on_demand_simple(definition):
    # given
    session = StubSession(is_open=True)
    response = definition.get_data(session)

    assert_create_df_on_demand(response)


@pytest.mark.parametrize(
    argnames="definition,response",
    ids=[
        "historical_pricing.events",
        "historical_pricing.summaries",
        "ipa.curves.forward_curves",
        "ipa.curves.zc_curve_definitions",
        "ipa.curves.zc_curves",
        "ipa.financial_contracts.bond",
        "ipa.financial_contracts.cap_floor",
        "ipa.financial_contracts.cds",
        "ipa.financial_contracts.cross",
        "ipa.financial_contracts.option",
        "ipa.financial_contracts.repo",
        "ipa.financial_contracts.swap",
        "ipa.financial_contracts.swaption",
        "ipa.financial_contracts.term_deposit",
        "ipa.surfaces.cap",
        "ipa.surfaces.eti",
        "ipa.surfaces.fx",
        "ipa.surfaces.swaption",
        "ipa.surfaces",
        "pricing",
    ],
    argvalues=[
        # <editor-fold desc="ipa">
        args(
            definition=rdc.ipa.curves.forward_curves.Definition(),
            response=FORWARD_CURVE_RESPONSE,
        ),
        args(
            definition=rdc.ipa.curves.forward_curves.Definitions([]),
            response=FORWARD_CURVES_RESPONSE,
        ),
        args(
            definition=rdc.ipa.curves.zc_curves.Definition(), response=ZC_CURVE_RESPONSE
        ),
        args(
            definition=rdc.ipa.curves.zc_curve_definitions.Definition(),
            response=ZC_CURVE_DEFINITION_RESPONSE,
        ),
        args(
            definition=rdc.ipa.curves.zc_curve_definitions.Definitions([]),
            response=ZC_CURVE_DEFINITIONS_RESPONSE,
        ),
        args(
            definition=rdc.ipa.curves.zc_curves.Definitions([]),
            response=ZC_CURVES_RESPONSE,
        ),
        args(
            definition=rdc.ipa.financial_contracts.bond.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.cap_floor.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.cds.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.cross.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.option.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.repo.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.swap.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.swaption.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(
            definition=rdc.ipa.financial_contracts.term_deposit.Definition(),
            response=StubResponse({"data": [], "headers": []}),
        ),
        args(definition=SURFACE_CAP_DEFINITION, response=SURFACE_CAP_RESPONSE),
        args(definition=SURFACE_ETI_DEFINITION, response=SURFACE_ETI_RESPONSE),
        args(definition=SURFACE_FX_DEFINITION, response=SURFACE_FX_RESPONSE),
        args(
            definition=SURFACE_SWAPTION_DEFINITION, response=SURFACE_SWAPTION_RESPONSE
        ),
        args(definition=SURFACES_DEFINITION, response=SURFACES_RESPONSE),
        # </editor-fold>
        # <editor-fold desc="pricing">
        args(definition=PRICING_DEFINITION, response=PRICING_RESPONSE),
        # </editor-fold>
    ],
)
def test_creating_df_on_demand_complex(definition, response):
    # given
    session = StubSession(is_open=True, response=response)
    response = definition.get_data(session)

    assert_create_df_on_demand(response)


@pytest.mark.parametrize(
    argnames="definition,response",
    ids=[
        "historical_pricing.events",
        "historical_pricing.summaries",
        "custom_instruments.events",
        "custom_instruments.summaries",
    ],
    argvalues=[
        # <editor-fold desc="historical_pricing">
        args(
            definition=rdc.historical_pricing.events.Definition("EUR"),
            response=StubResponse(
                [{"headers": [{"name": "DATE"}], "data": [{}], "universe": {"ric": ""}}]
            ),
        ),
        args(
            definition=rdc.historical_pricing.summaries.Definition("EUR"),
            response=StubResponse(
                [{"headers": [{"name": "DATE"}], "data": [{}], "universe": {"ric": ""}}]
            ),
        ),
        # </editor-fold>
        # <editor-fold desc="custom_instruments">
        args(
            definition=rdc.custom_instruments.events.Definition("VOD.L"),
            response=StubResponse(
                [{"headers": [{"name": "DATE"}], "data": [{}], "universe": {"ric": ""}}]
            ),
        ),
        args(
            definition=rdc.custom_instruments.summaries.Definition("VOD.L"),
            response=StubResponse(
                [{"headers": [{"name": "DATE"}], "data": [{}], "universe": {"ric": ""}}]
            ),
        ),
        # </editor-fold>
    ],
)
def test_creating_df_on_demand_complex(definition, response):
    # given
    session = StubSession(is_open=True, response=response)
    response = definition.get_data(session)

    assert_create_df_on_demand(response)


@pytest.mark.parametrize(
    argnames="definition,response",
    ids=[
        "fundamental_and_reference",
        "news.headlines",
    ],
    argvalues=[
        args(
            definition=FUNDAMENTAL_AND_REFERENCE_DEFINITION,
            response=FUNDAMENTAL_AND_REFERENCE_RESPONSE_PLATFORM,
        ),
        args(definition=NEWS_HEADLINES_DEFINITION, response=NEWS_HEADLINES_RESPONSE),
    ],
)
def test_creating_df_on_demand_rdp(definition, response):
    # given
    platform = "rdp"
    config = StubConfig(
        {
            "apis.data.datagrid": {
                "url": "/data/datagrid/beta1",
                "underlying-platform": platform,
                "endpoints.standard": "/",
            },
            "apis.data.news": {"endpoints.headlines": "headlines/url", "url": "url"},
            "apis.data.news.underlying-platform": platform,
            "raise_exception_on_error": True,
        }
    )
    session = StubSession(is_open=True, response=response, config=config)
    response = definition.get_data(session=session)

    assert_create_df_on_demand(response)


@pytest.mark.parametrize(
    argnames="definition,response",
    ids=[
        "fundamental_and_reference",
        "news.headlines",
    ],
    argvalues=[
        args(
            definition=FUNDAMENTAL_AND_REFERENCE_DEFINITION,
            response=FUNDAMENTAL_AND_REFERENCE_RESPONSE_DESKTOP,
        ),
        args(
            definition=NEWS_HEADLINES_DEFINITION, response=NEWS_HEADLINES_RESPONSE_UDF
        ),
    ],
)
def test_creating_df_on_demand_udf(definition, response):
    # given
    platform = "udf"
    config = StubConfig(
        {
            "apis.data.datagrid": {
                "url": "/data/datagrid/beta1",
                "underlying-platform": platform,
                "endpoints.standard": "/",
            },
            "apis.data.news": {"endpoints.headlines": "headlines/url", "url": "url"},
            "apis.data.news.underlying-platform": platform,
            "raise_exception_on_error": True,
        }
    )
    session = StubSession(is_open=True, response=response, config=config)
    session.type = SessionType.DESKTOP
    response = definition.get_data(session=session)

    assert_create_df_on_demand(response)



import allure
import pytest

from refinitiv.data.content.ownership._enums import Frequency
from refinitiv.data.content.ownership.fund import shareholders_history_report
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "TR.FundPortfolioName",
    "TR.FundParentType",
    "TR.FundClassId",
    "TR.FundInvestorType",
    "TR.FundRegion",
    "TR.FundInvtStyleCode",
    "TR.FundInvtOrientation",
    "TR.FundTurnoverRating",
    "TR.FundOwnershipTurnover",
    "TR.FundTotalEquityAssets",
    "TR.FundAdjShrsHeld",
    "TR.FdAdjPctOfShrsOutHeld",
    "TR.FdAdjSharesHeldValue",
    "TR.FundHoldingsDate",
    "TR.FundFilingType",
    "TR.FdAdjShrsHeldChange",
    "TR.FdAdjShrsHeldValChg",
    "TR.PrevFundAdjSharesHeld",
    "TR.PrevAdjSharesHeldValue",
    "TR.PrevAdjPctShrsOutHeld",
    "TR.PrevFundHoldingsDate",
    "TR.PrevFundFilingType",
]

EXPECTED_COLUMN_TITLES = [
    "Instrument",
    "Investor Perm Id",
    "Investor Name",
    "Investor Parent Type",
    "Fund ClassId",
    "Investor Type",
    "Investor Region",
    "Investment Style",
    "Orientation",
    "Turnover Rating",
    "Turnover %",
    "Total Equity Assets",
    "SharesHeld",
    "% SharesOutstanding",
    "SharesHeld Value",
    "Holdings Date",
    "Filing Type",
    "SharesHeld Change",
    "SharesHeld Value Change",
    "Previous SharesHeld",
    "Previous SharesHeld Value",
    "Previous % SharesOutstanding",
    "Previous Holdings Date",
    "Previous Filing Type",
]


@allure.suite("Content object - Ownership Fund Shareholders History Report")
@allure.feature("Content object - Ownership Fund Shareholders History Report")
@allure.severity(allure.severity_level.CRITICAL)
class TestFundShareholdersHistoryReport:
    @allure.title(
        "Create Fund Shareholders History Report definition object with valid required params and get data"
    )
    @pytest.mark.caseid("37478943")
    @pytest.mark.parametrize(
        "universe,frequency,extended_params,",
        [
            ("ONEX.CCP", "M", {"test_param": "test_value"}),
        ],
    )
    @pytest.mark.smoke
    def test_create_fund_shareholders_history_report_object_with_valid_fields_and_get_data(
        self,
        open_platform_session,
        universe,
        frequency,
        extended_params,
    ):
        response = shareholders_history_report.Definition(
            universe=universe,
            frequency=frequency,
            extended_params=extended_params,
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(
            response, EXPECTED_COLUMN_TITLES
        )
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )
        check_dataframe_column_date_for_datetime_type(response)
        check_extended_params_were_sent_in_request(response, extended_params)

    @allure.title(
        "Create Fund Shareholders History Report definition object with all valid params and get data asynchronously"
    )
    @pytest.mark.caseid("37478944")
    @pytest.mark.parametrize(
        "universe,frequency,start,end, limit",
        [("ONEX.CCP", Frequency.QUARTERLY, "20210925", "20211005", 4)],
    )
    async def test_create_fund_shareholders_history_report_object_with_valid_fields_and_get_data_async(
        self, open_platform_session_async, universe, frequency, start, end, limit
    ):
        response = await get_async_response_from_definition(
            shareholders_history_report.Definition(
                universe, frequency, start, end, limit, use_field_names_in_headers=True
            )
        )
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, EXPECTED_COLUMN_NAMES)
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )



from refinitiv.data._core.session import Sensitive, GrantPassword


def test_sensitive():
    # given
    expected = "psadfjg@341F32"

    # when
    password = Sensitive("psadfjg@341F32")

    # then
    assert expected == password
    assert repr(password) == "********"


def test_password_is_sensitive():
    # given
    # when
    grant = GrantPassword("username", "password")

    # then
    assert isinstance(grant._password, Sensitive)



from typing import TYPE_CHECKING

from ....._object_definition import ObjectDefinition
from ..._enums import MainConstituentAssetClass, RiskType


if TYPE_CHECKING:
    from ..._types import OptMainConstituentAssetClass, OptRiskType
    from ......._types import OptStr, OptBool


class CrossCurrencyCurveGetDefinitionItem(ObjectDefinition):
    """
    Returns the definitions of the available commodity and energy forward curves for
    the filters selected (e.g. currency, source…).

    Parameters
    ----------
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    id : str, optional
        The identifier of the cross currency definitions.
    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    valuation_date : str, optional
        The date used to define a list of curves or a unique cross currency curve that
        can be priced at this date. the value is expressed in iso 8601 format:
        yyyy-mm-dd (e.g., '2021-01-01').
    """

    def __init__(
        self,
        main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
        risk_type: "OptRiskType" = None,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        curve_tag: "OptStr" = None,
        id: "OptStr" = None,
        is_non_deliverable: "OptBool" = None,
        name: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        source: "OptStr" = None,
        valuation_date: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.main_constituent_asset_class = main_constituent_asset_class
        self.risk_type = risk_type
        self.base_currency = base_currency
        self.base_index_name = base_index_name
        self.curve_tag = curve_tag
        self.id = id
        self.is_non_deliverable = is_non_deliverable
        self.name = name
        self.quoted_currency = quoted_currency
        self.quoted_index_name = quoted_index_name
        self.source = source
        self.valuation_date = valuation_date

    @property
    def main_constituent_asset_class(self):
        """
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
        :return: enum MainConstituentAssetClass
        """
        return self._get_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass"
        )

    @main_constituent_asset_class.setter
    def main_constituent_asset_class(self, value):
        self._set_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass", value
        )

    @property
    def risk_type(self):
        """
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
        :return: enum RiskType
        """
        return self._get_enum_parameter(RiskType, "riskType")

    @risk_type.setter
    def risk_type(self, value):
        self._set_enum_parameter(RiskType, "riskType", value)

    @property
    def base_currency(self):
        """
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
        :return: str
        """
        return self._get_parameter("baseCurrency")

    @base_currency.setter
    def base_currency(self, value):
        self._set_parameter("baseCurrency", value)

    @property
    def base_index_name(self):
        """
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
        :return: str
        """
        return self._get_parameter("baseIndexName")

    @base_index_name.setter
    def base_index_name(self, value):
        self._set_parameter("baseIndexName", value)

    @property
    def curve_tag(self):
        """
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)

    @property
    def id(self):
        """
        The identifier of the cross currency definitions.
        :return: str
        """
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)

    @property
    def is_non_deliverable(self):
        """
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
        :return: bool
        """
        return self._get_parameter("isNonDeliverable")

    @is_non_deliverable.setter
    def is_non_deliverable(self, value):
        self._set_parameter("isNonDeliverable", value)

    @property
    def name(self):
        """
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
        :return: str
        """
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)

    @property
    def quoted_currency(self):
        """
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
        :return: str
        """
        return self._get_parameter("quotedCurrency")

    @quoted_currency.setter
    def quoted_currency(self, value):
        self._set_parameter("quotedCurrency", value)

    @property
    def quoted_index_name(self):
        """
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
        :return: str
        """
        return self._get_parameter("quotedIndexName")

    @quoted_index_name.setter
    def quoted_index_name(self, value):
        self._set_parameter("quotedIndexName", value)

    @property
    def source(self):
        """
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
        :return: str
        """
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)

    @property
    def valuation_date(self):
        """
        The date used to define a list of curves or a unique cross currency curve that
        can be priced at this date. the value is expressed in iso 8601 format:
        yyyy-mm-dd (e.g., '2021-01-01').
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf8

from .._object_definition import ObjectDefinition
from ...._tools import try_copy_to_list
from ...._types import OptStr, OptStrings


class ForwardCurveDefinition(ObjectDefinition):
    """
    Parameters
    ----------
    index_tenor : str, optional

    forward_curve_tenors : list of str, optional
        Defines expected forward rate surface tenor/slices
    forward_curve_tag : str, optional

    forward_start_date : str, optional
        Defines the forward start date by date format
    forward_start_tenor : str, optional
        Defines the forward start date by tenor format: "Spot" / "1M" / ...
    """

    def __init__(
        self,
        index_tenor: OptStr = None,
        forward_curve_tag: OptStr = None,
        forward_curve_tenors: OptStrings = None,
        forward_start_date: OptStr = None,
        forward_start_tenor: OptStr = None,
    ) -> None:
        super().__init__()
        self.index_tenor = index_tenor
        self.forward_curve_tenors = try_copy_to_list(forward_curve_tenors)
        self.forward_curve_tag = forward_curve_tag
        self.forward_start_date = forward_start_date
        self.forward_start_tenor = forward_start_tenor

    @property
    def forward_curve_tenors(self):
        """
        Defines expected forward rate surface tenor/slices
        :return: list string
        """
        return self._get_list_parameter(str, "forwardCurveTenors")

    @forward_curve_tenors.setter
    def forward_curve_tenors(self, value):
        self._set_list_parameter(str, "forwardCurveTenors", value)

    @property
    def forward_curve_tag(self):
        """
        :return: str
        """
        return self._get_parameter("forwardCurveTag")

    @forward_curve_tag.setter
    def forward_curve_tag(self, value):
        self._set_parameter("forwardCurveTag", value)

    @property
    def forward_start_date(self):
        """
        Defines the forward start date by date format
        :return: str
        """
        return self._get_parameter("forwardStartDate")

    @forward_start_date.setter
    def forward_start_date(self, value):
        self._set_parameter("forwardStartDate", value)

    @property
    def forward_start_tenor(self):
        """
        Defines the forward start date by tenor format: "Spot" / "1M" / ...
        :return: str
        """
        return self._get_parameter("forwardStartTenor")

    @forward_start_tenor.setter
    def forward_start_tenor(self, value):
        self._set_parameter("forwardStartTenor", value)

    @property
    def index_tenor(self):
        """
        :return: str
        """
        return self._get_parameter("indexTenor")

    @index_tenor.setter
    def index_tenor(self, value):
        self._set_parameter("indexTenor", value)



import pytest

from refinitiv.data._tools import merge_dict_to_dict
from refinitiv.data._tools._utils import camel_to_snake
from tests.unit.conftest import args


@pytest.mark.parametrize(
    "input_str, expected",
    [
        ("", ""),
        ("CamelCase", "camel_case"),
        ("camelCamelCase", "camel_camel_case"),
        ("camel2_camel2_case", "camel2_camel2_case"),
        ("getHTTPResponseCode", "get_http_response_code"),
        ("get2HTTPResponse123Code", "get2_http_response123_code"),
        ("HTTPResponseCodeXYZ", "http_response_code_xyz"),
        ("10CoolDudes", "10_cool_dudes"),
    ],
)
def test_base_create_str_definition_without_correct_module_path(input_str, expected):
    result = camel_to_snake(input_str)

    assert result == expected


@pytest.mark.parametrize(
    "dest, source, expected",
    [
        args(dest={}, source={}, expected={}),
        args(dest={"key": "value"}, source={"key": "value"}, expected={"key": "value"}),
        args(
            dest={"key": "value_old"},
            source={"key": "value_new"},
            expected={"key": "value_new"},
        ),
        args(
            dest={"key_old": "value_old"},
            source={"key_new": "value_new"},
            expected={"key_old": "value_old", "key_new": "value_new"},
        ),
        args(dest={"key": 1}, source={"key": 2}, expected={"key": 2}),
        args(
            dest={"key": [1, 2, 3]},
            source={"key": [4, 5, 6]},
            expected={"key": [4, 5, 6]},
        ),
        args(
            dest={"key1": {"key2": "value2"}},
            source={"key1": {"key2": "value_new", "key3": "value3"}},
            expected={"key1": {"key2": "value_new", "key3": "value3"}},
        ),
    ],
)
def test_merge_dict_to_dict(dest, source, expected):
    testing = merge_dict_to_dict(dest, source)
    assert testing == expected



import re
from typing import List, Union, TYPE_CHECKING

import numpy as np
import pandas as pd
from pandas import DataFrame, MultiIndex
from pandas.core.dtypes.cast import convert_dtypes as pandas_covert_dtypes
from pandas.core.dtypes.common import (
    is_object_dtype,
    is_datetime64_any_dtype,
    pandas_dtype,
)
from pandas.core.tools.datetimes import DatetimeScalarOrArrayConvertible


DtypeObj = Union[np.dtype, "ExtensionDtype"]


if TYPE_CHECKING:
    from pandas.core.series import Series


def convert_df_columns_to_datetime(
    df: pd.DataFrame, entry: str, utc: bool = None, delete_tz: bool = False
) -> pd.DataFrame:
    """Converts particular dataframe columns to datetime according the pattern.

    Converts particular dataframe column or columns if one of more columns
    matches the pattern, returns same dataframe otherwise.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas dataframe to convert.
    entry: str
        Pattern to find a column to convert.
    utc : bool
        Convert to UTC if True.
    delete_tz : bool
        Convert to timezone-unaware if True.

    Returns
    -------
    pd.DataFrame
        Converted dataframe
    """
    columns_indexes = [
        index
        for index, name in enumerate(df.columns.values)
        if entry.lower() in name.lower()
    ]

    return convert_df_columns_to_datetime_by_idx(df, columns_indexes, utc, delete_tz)


def convert_df_columns_to_datetime_by_idx(
    df: pd.DataFrame,
    columns_indexes: List[int],
    utc: bool = None,
    delete_tz: bool = False,
):
    """Convert dataframe columns to datetime by index.

    Parameters
    ----------
    df : pd.Dataframe
        Pandas dataframe to convert.
    columns_indexes : List[int]
        List of indexes of columns to convert.
    utc : bool
        Convert to UTC if True.
    delete_tz : bool
        Convert to timezone-unaware if True.

    Returns
    -------
    df
        Converted dataframe.
    """
    for idx in columns_indexes:
        df.iloc[:, idx] = pd.to_datetime(df.iloc[:, idx], utc=utc, errors="coerce")
        if delete_tz:
            df.iloc[:, idx] = df.iloc[:, idx].dt.tz_localize(None)

    return df


def convert_df_columns_to_datetime_re(
    df: pd.DataFrame, pattern: re.compile
) -> pd.DataFrame:
    """Convert dataframe columns to datetime using regular expression pattern.

    Parameters
    ----------
    df : pd.Dataframe
        Pandas dataframe to convert.
    pattern : re.compile
        Regular expression pattern to check columns.

    Returns
    -------
    df
        Converted dataframe
    """
    column_indexes = [
        index for index, name in enumerate(df.columns.values) if pattern.search(name)
    ]

    return convert_df_columns_to_datetime_by_idx(
        df, column_indexes, utc=True, delete_tz=True
    )


def convert_str_to_datetime(s: str) -> DatetimeScalarOrArrayConvertible:
    date = pd.to_datetime(s, utc=True, errors="coerce")
    date = date.tz_localize(None)
    return date


def sort_df_by_universe(df: DataFrame, universe: List[str]) -> DataFrame:
    length = len(universe)

    if length == 1:
        return df

    columns = df.columns

    def make_getidx():
        get_index = universe.index
        if isinstance(columns, MultiIndex):

            def geti(i):
                return i[0]

        else:

            def geti(i):
                return i

        def inner(i):
            try:
                index = get_index(geti(i))
            except ValueError:
                index = length
            return index

        return inner

    getidx = make_getidx()
    # [3, 0, 2, 1]
    curr_order = [getidx(col) for col in columns]
    # [0, 1, 2, 3]
    expected_order = list(range(length))
    if curr_order != expected_order:
        sorted_columns = (col for _, col in sorted(zip(curr_order, columns)))
        df = df.reindex(columns=sorted_columns)
    return df


def convert_dtypes(df: pd.DataFrame) -> pd.DataFrame:
    """
    This function is an extension to the standard pandas.DataFrame.convert_dtypes.

    Correct return dataframe if we have this columns in dataframe:

    GOOG.O                    Currency
    Date
    2020-12-31 00:00:00+00:00     <NA>
    2020-12-31 00:00:00+00:00     <NA>

    Correct convert None, np.nan, pd.NA, pd.NaN to pd.NA, see official docs:
    https://pandas.pydata.org/pandas-docs/stable/user_guide/missing_data.html#missing-data-na

    Correct convert big int from Linux, Windows platform.


    Parameters
    ----------
    df: pd.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    if df.empty:
        return df

    df.fillna(np.nan, inplace=True)
    columns_indexes = [index for index, _ in enumerate(df.columns.values)]

    for index in columns_indexes:
        series = df.iloc[:, index]
        if is_datetime64_any_dtype(series.dtype):
            continue

        series = series.infer_objects()
        if is_object_dtype(series):
            series = series.copy()

        inferred_dtype = _get_inferred_dtype(series)
        if str(inferred_dtype).lower() == "object":
            new_series = series.copy()
            new_series.replace("", np.nan, inplace=True)

            inferred_dtype = _get_inferred_dtype(new_series)
            if str(inferred_dtype).lower() != "object":
                series = new_series

        result = series.astype(inferred_dtype)
        if hasattr(df, "isetitem"):
            df.isetitem(index, result)
        else:
            df.iloc[:, index] = result

    df.fillna(pd.NA, inplace=True)
    return df


def _get_inferred_dtype(series: "Series") -> DtypeObj:
    inferred_dtype = pandas_covert_dtypes(series.values)
    if "float" in str(inferred_dtype).lower() and (series.fillna(-9999) % 1 == 0).all():
        return pandas_dtype("Int64")
    return inferred_dtype
